#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import copy
from unittest import mock
import uuid

import netaddr
from neutron_lib import constants
from oslo_utils import netutils
from oslo_utils import uuidutils
from ovsdbapp.backend.ovs_idl import connection
from ovsdbapp import constants as const
from ovsdbapp import event as ovsdb_event
from ovsdbapp.tests.functional import base
from ovsdbapp.tests import utils

from neutron.common.ovn import constants as ovn_const
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb \
    import impl_idl_ovn as impl
from neutron.services.portforwarding import constants as pf_const
from neutron.tests.functional import base as n_base
from neutron.tests.functional.common import ovn as ovn_common
from neutron.tests.functional.resources.ovsdb import events

OWNER = ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY
PF_PLUGIN = pf_const.PORT_FORWARDING_PLUGIN


class BaseOvnIdlTest(n_base.BaseLoggingTestCase,
                     base.FunctionalTestCase):
    schemas = ['OVN_Southbound', 'OVN_Northbound']

    def setUp(self):
        super(BaseOvnIdlTest, self).setUp()
        ovn_conf.register_opts()
        self.api = impl.OvsdbSbOvnIdl(self.connection['OVN_Southbound'])
        self.nbapi = impl.OvsdbNbOvnIdl(self.connection['OVN_Northbound'])
        self.handler = ovsdb_event.RowEventHandler()
        self.api.idl.notify = self.handler.notify


class TestSbApi(BaseOvnIdlTest):

    def setUp(self):
        super(TestSbApi, self).setUp()
        self.data = {
            'chassis': [
                {'other_config': {'ovn-bridge-mappings':
                                  'public:br-ex,private:br-0'}},
                {'other_config': {'ovn-bridge-mappings':
                                  'public:br-ex,public2:br-ex2'}},
                {'other_config': {'ovn-bridge-mappings':
                                  'public:br-ex'}},
            ]
        }
        self.load_test_data()

    def load_test_data(self):
        with self.api.transaction(check_error=True) as txn:
            for chassis in self.data['chassis']:
                chassis['name'] = utils.get_rand_device_name('chassis')
                chassis['hostname'] = '%s.localdomain.com' % chassis['name']
                txn.add(self.api.chassis_add(
                    chassis['name'], ['geneve'], chassis['hostname'],
                    hostname=chassis['hostname'],
                    other_config=chassis['other_config']))

    def test_get_chassis_hostname_and_physnets(self):
        mapping = self.api.get_chassis_hostname_and_physnets()
        self.assertLessEqual(len(self.data['chassis']), len(mapping))
        self.assertGreaterEqual(set(mapping.keys()),
                                {c['hostname'] for c in self.data['chassis']})

    def test_get_all_chassis(self):
        chassis_list = set(self.api.get_all_chassis())
        our_chassis = {c['name'] for c in self.data['chassis']}
        self.assertLessEqual(our_chassis, chassis_list)

    def test_chassis_exists(self):
        self.assertTrue(self.api.chassis_exists(
            self.data['chassis'][0]['hostname']))
        self.assertFalse(self.api.chassis_exists("nochassishere"))

    def test_get_chassis_and_physnets(self):
        mapping = self.api.get_chassis_and_physnets()
        self.assertLessEqual(len(self.data['chassis']), len(mapping))
        self.assertGreaterEqual(set(mapping.keys()),
                                {c['name'] for c in self.data['chassis']})

    def test_multiple_physnets_in_one_bridge(self):
        self.data = {
            'chassis': [
                {'other_config': {'ovn-bridge-mappings': 'p1:br-ex,p2:br-ex'}}
            ]
        }
        self.load_test_data()
        self.assertRaises(ValueError, self.api.get_chassis_and_physnets)

    def _add_switch(self, chassis_name):
        sname = utils.get_rand_device_name(prefix='switch')
        chassis = self.api.lookup('Chassis', chassis_name)
        with self.nbapi.transaction(check_error=True) as txn:
            switch = txn.add(self.nbapi.ls_add(sname))
        return chassis, switch.result

    def _add_port_to_switch(
            self, switch, type=ovn_const.LSP_TYPE_LOCALPORT,
            device_owner=constants.DEVICE_OWNER_DISTRIBUTED):
        pname = utils.get_rand_device_name(prefix='port')
        row_event = events.WaitForCreatePortBindingEvent(pname)
        self.handler.watch_event(row_event)
        with self.nbapi.transaction(check_error=True) as txn:
            port = txn.add(self.nbapi.lsp_add(
                switch.uuid, pname, type=type,
                external_ids={
                    ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY: device_owner}))
        row_event.wait()
        return port.result, row_event.row

    def test_get_metadata_port_network(self):
        chassis, switch = self._add_switch(self.data['chassis'][0]['name'])
        port, binding = self._add_port_to_switch(switch)
        result = self.api.get_metadata_port_network(str(binding.datapath.uuid))
        self.assertEqual(binding, result)
        self.assertEqual(binding.datapath.external_ids['logical-switch'],
                         str(switch.uuid))
        self.assertEqual(
            port.external_ids[ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY],
            constants.DEVICE_OWNER_DISTRIBUTED)

    def test_get_metadata_port_network_other_non_metadata_port(self):
        chassis, switch = self._add_switch(self.data['chassis'][0]['name'])
        port, binding = self._add_port_to_switch(switch)
        port_lbhm, binding_port_lbhm = self._add_port_to_switch(
            switch, device_owner=ovn_const.OVN_LB_HM_PORT_DISTRIBUTED)
        result = self.api.get_metadata_port_network(str(binding.datapath.uuid))
        self.assertEqual(binding, result)
        self.assertEqual(binding.datapath.external_ids['logical-switch'],
                         str(switch.uuid))
        self.assertEqual(
            binding_port_lbhm.datapath.external_ids['logical-switch'],
            str(switch.uuid))
        self.assertEqual(
            port_lbhm.external_ids[ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY],
            ovn_const.OVN_LB_HM_PORT_DISTRIBUTED)

    def test_get_metadata_port_network_missing(self):
        val = str(uuid.uuid4())
        self.assertIsNone(self.api.get_metadata_port_network(val))

    def _create_bound_port_with_ip(self, mac, ipaddr):
        chassis, switch = self._add_switch(
            self.data['chassis'][0]['name'])
        port, binding = self._add_port_to_switch(switch)
        mac_ip = '%s %s' % (mac, ipaddr)
        pb_update_event = events.WaitForUpdatePortBindingEvent(
            port.name, mac=[mac_ip])
        self.handler.watch_event(pb_update_event)
        self.nbapi.lsp_set_addresses(
            port.name, [mac_ip]).execute(check_error=True)
        self.assertTrue(pb_update_event.wait())
        self.api.lsp_bind(port.name, chassis.name).execute(check_error=True)

        return binding, switch

    def test_get_network_port_bindings_by_ip(self):
        mac = 'de:ad:be:ef:4d:ad'
        ipaddr = '192.0.2.1'
        binding, switch = self._create_bound_port_with_ip(mac, ipaddr)
        # binding, ipaddr, switch = self._create_bound_port_with_ip()
        network_id = switch.name.replace('neutron-', '')
        result = self.api.get_network_port_bindings_by_ip(network_id, ipaddr)
        self.assertIn(binding, result)

    def test_get_network_port_bindings_by_ip_ipv6_ll(self):
        ipaddr = 'fe80::99'
        mac = str(netutils.get_mac_addr_by_ipv6(netaddr.IPAddress(ipaddr)))
        binding, switch = self._create_bound_port_with_ip(mac, ipaddr)
        network_id = switch.name.replace('neutron-', '')
        result = self.api.get_network_port_bindings_by_ip(network_id, ipaddr)
        self.assertIn(binding, result)

    def test_get_network_port_bindings_by_ip_with_unbound_port(self):
        mac = 'de:ad:be:ef:4d:ad'
        ipaddr = '192.0.2.1'
        binding, switch = self._create_bound_port_with_ip(mac, ipaddr)
        unbound_port_name = utils.get_rand_device_name(prefix="port")
        mac_ip = "de:ad:be:ef:4d:ab %s" % ipaddr
        with self.nbapi.transaction(check_error=True) as txn:
            txn.add(
                self.nbapi.lsp_add(switch.name, unbound_port_name, type=type))
            txn.add(self.nbapi.lsp_set_addresses(unbound_port_name, [mac_ip]))
        network_id = switch.name.replace('neutron-', '')
        result = self.api.get_network_port_bindings_by_ip(network_id, ipaddr)
        self.assertIn(binding, result)
        self.assertEqual(1, len(result))

    def test_get_ports_on_chassis(self):
        chassis, switch = self._add_switch(
            self.data['chassis'][0]['name'])
        port, binding = self._add_port_to_switch(switch)
        self.api.lsp_bind(port.name, chassis.name).execute(check_error=True)
        self.assertEqual([binding],
                         self.api.get_ports_on_chassis(chassis.name))

    def _test_get_ports_on_chassis_with_additional_chassis(
            self, ports, chassis, bindings, expected):
        self.api.lsp_bind(
            ports[0].name, chassis[0].name).execute(check_error=True)
        self.api.lsp_bind(
            ports[1].name, chassis[1].name).execute(check_error=True)

        self.api.db_set('Port_Binding', bindings[1].uuid,
            additional_chassis=[chassis[0].uuid]).execute(
                check_error=True, log_errors=True)

        result = self.api.get_ports_on_chassis(
            chassis[0].name, include_additional_chassis=True)

        self.assertEqual(expected, result)

    @ovn_common.skip_if_additional_chassis_not_supported('api')
    def test_get_ports_on_chassis_with_additional_chassis(self):
        chassis, switch = self._add_switch(self.data['chassis'][0]['name'])
        port, binding = self._add_port_to_switch(switch)
        chassis2, switch2 = self._add_switch(self.data['chassis'][1]['name'])
        port2, binding2 = self._add_port_to_switch(switch2)

        self._test_get_ports_on_chassis_with_additional_chassis(
            ports=[port, port2],
            chassis=[chassis, chassis2],
            bindings=[binding, binding2],
            expected=[binding, binding2])

    def test_get_ports_on_chassis_with_additional_chassis_not_supported(self):
        chassis, switch = self._add_switch(self.data['chassis'][0]['name'])
        port, binding = self._add_port_to_switch(switch)
        chassis2, switch2 = self._add_switch(self.data['chassis'][1]['name'])
        port2, binding2 = self._add_port_to_switch(switch2)

        with mock.patch(
                'neutron.common.ovn.utils.is_additional_chassis_supported',
                return_value=False):
            self._test_get_ports_on_chassis_with_additional_chassis(
                ports=[port, port2],
                chassis=[chassis, chassis2],
                bindings=[binding, binding2],
                expected=[binding])


class TestNbApi(BaseOvnIdlTest):

    def setUp(self):
        super(TestNbApi, self).setUp()
        self.data = {
            'lbs': [
                {'name': 'pf-floatingip-fip_id1-tcp',
                 'protocol': const.PROTO_TCP,
                 'vips': {"172.24.4.8:2020": ["10.0.0.10:22"],
                          "172.24.4.8:2021": ["10.0.0.11:22"]},
                 'external_ids': {
                     ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY:
                         pf_const.PORT_FORWARDING_PLUGIN,
                     ovn_const.OVN_FIP_EXT_ID_KEY: 'fip_id1',
                     ovn_const.OVN_REV_NUM_EXT_ID_KEY: '1',
                     ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY: 'neutron-rtr1_id'}},
                {'name': 'pf-floatingip-fip_id1-udp',
                 'protocol': const.PROTO_UDP,
                 'vips': {"172.24.4.8:53": ["10.0.0.10:53"]},
                 'external_ids': {
                     ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY:
                         pf_const.PORT_FORWARDING_PLUGIN,
                     ovn_const.OVN_FIP_EXT_ID_KEY: 'fip_id1',
                     ovn_const.OVN_REV_NUM_EXT_ID_KEY: '1',
                     ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY: 'neutron-rtr1_id'}},
                {'name': 'pf-floatingip-fip_id2-tcp',
                 'protocol': const.PROTO_TCP,
                 'vips': {"172.24.4.100:2020": ["10.0.0.10:22"]},
                 'external_ids': {
                     ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY:
                         pf_const.PORT_FORWARDING_PLUGIN,
                     ovn_const.OVN_FIP_EXT_ID_KEY: 'fip_id2',
                     ovn_const.OVN_REV_NUM_EXT_ID_KEY: '10',
                     ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY: 'neutron-rtr2_id'}},
                {'name': 'octavia_lb1_id',
                 'vips': {"172.24.4.250:80": ["10.0.0.10:8080",
                                              "10.0.0.11:8080"]},
                 'selection_fields': ['ip_dst', 'ip_src', 'tp_dst', 'tp_src'],
                 'external_ids': {
                     'enabled': 'True',
                     'lr_ref': 'neutron-rtr1_id',
                     'ls_refs': str({'neutron-net1_id': 1}),
                     ovn_const.LB_EXT_IDS_VIP_KEY: '172.24.4.250',
                     ovn_const.LB_EXT_IDS_VIP_PORT_ID_KEY: 'neutron-port1_id',
                 }},
            ],
        }
        self.load_test_data()

    def load_test_data(self):
        with self.nbapi.transaction(check_error=True) as txn:
            for lb in copy.deepcopy(self.data['lbs']):
                lb_name = lb.pop('name')
                for vip, ips in lb.pop('vips').items():
                    txn.add(self.nbapi.lb_add(
                        lb_name, vip, ips, may_exist=True, **lb))

    def test_lb_list(self):
        lbs = self.nbapi.lb_list().execute(check_error=True)
        self.assertEqual(len(self.data['lbs']), len(lbs))
        exp_values = [(lb['name'], lb['external_ids'])
                      for lb in self.data['lbs']]
        lbs_values = [(lb.name, lb.external_ids) for lb in lbs]
        self.assertCountEqual(exp_values, lbs_values)

    def test_get_router_floatingip_lbs(self):
        f = self.nbapi.get_router_floatingip_lbs
        self.assertEqual([], f('unused_router_name'))
        for exp_router_name in ['neutron-rtr1_id', 'neutron-rtr2_id']:
            exp_values = [
                (lb['name'], lb['external_ids'])
                for lb in self.data['lbs']
                if all([lb['external_ids'].get(OWNER) == PF_PLUGIN,
                        lb['external_ids'].get(
                            ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY) ==
                        exp_router_name])]
            lbs_values = [(lb.name, lb.external_ids)
                          for lb in f(exp_router_name)]
            self.assertTrue(exp_values)
            self.assertCountEqual(exp_values, lbs_values)

    def test_get_floatingip_in_nat_or_lb(self):
        f = self.nbapi.get_floatingip_in_nat_or_lb
        self.assertIsNone(f('unused_fip_id'))
        for exp_fip_id in ['fip_id1', 'fip_id2']:
            exp_values = [
                (lb['name'], lb['external_ids'])
                for lb in self.data['lbs']
                if all([lb['external_ids'].get(OWNER) == PF_PLUGIN,
                        lb['external_ids'].get(
                            ovn_const.OVN_FIP_EXT_ID_KEY) == exp_fip_id])]
            # get_floatingip_in_nat_or_lb returns the first entry
            # it finds for the fip_id provided. Make sure that it
            # is present in exp_values.
            lb_match = f(exp_fip_id)
            self.assertIn((lb_match['name'], lb_match['external_ids']),
                          exp_values)

    def test_create_lswitch_port_ha_chassis_group(self):
        ls_name = uuidutils.generate_uuid()
        lsp_name = uuidutils.generate_uuid()
        hcg_name = uuidutils.generate_uuid()
        self.nbapi.ha_chassis_group_add(hcg_name).execute(check_error=True)
        hcg = self.nbapi.lookup('HA_Chassis_Group', hcg_name)
        self.nbapi.ls_add(ls_name).execute(check_error=True)
        self.nbapi.create_lswitch_port(
            lsp_name, ls_name, ha_chassis_group=hcg.uuid).execute(
            check_error=True)
        lsp = self.nbapi.lookup('Logical_Switch_Port', lsp_name)
        self.assertEqual(hcg.uuid, lsp.ha_chassis_group[0].uuid)

    def test_set_lswitch_port_ha_chassis_group(self):
        ls_name = uuidutils.generate_uuid()
        lsp_name = uuidutils.generate_uuid()
        self.nbapi.ls_add(ls_name).execute(check_error=True)
        self.nbapi.create_lswitch_port(lsp_name, ls_name).execute(
            check_error=True)
        lsp = self.nbapi.lookup('Logical_Switch_Port', lsp_name)
        self.assertEqual([], lsp.ha_chassis_group)

        # Create an HA Chassis Group register and assign to the LSP.
        hcg_name = uuidutils.generate_uuid()
        self.nbapi.ha_chassis_group_add(hcg_name).execute(check_error=True)
        hcg = self.nbapi.lookup('HA_Chassis_Group', hcg_name)
        self.nbapi.set_lswitch_port(
            lsp_name, ha_chassis_group=hcg.uuid).execute(check_error=True)
        lsp = self.nbapi.lookup('Logical_Switch_Port', lsp_name)
        self.assertEqual(hcg.uuid, lsp.ha_chassis_group[0].uuid)

        # Unassign the HA Chassis Group from the LSP.
        self.nbapi.set_lswitch_port(
            lsp_name, ha_chassis_group=[]).execute(check_error=True)
        lsp = self.nbapi.lookup('Logical_Switch_Port', lsp_name)
        self.assertEqual([], lsp.ha_chassis_group)

    def test_set_router_mac_aging(self):
        name = "aging_router1"
        with self.nbapi.transaction(check_error=True) as txn:
            r = txn.add(self.nbapi.lr_add(name))
            txn.add(self.nbapi.set_router_mac_age_limit(router=name))
        self.assertEqual(r.result.options[ovn_const.LR_OPTIONS_MAC_AGE_LIMIT],
                         ovn_conf.get_ovn_mac_binding_age_threshold())

    def test_set_router_mac_aging_all(self):
        ovn_conf.cfg.CONF.set_override("mac_binding_age_threshold", 5,
                                       group="ovn")
        names = ["aging_router2", "aging_router3"]
        with self.nbapi.transaction(check_error=True) as txn:
            for name in names:
                txn.add(self.nbapi.lr_add(name))
            txn.add(self.nbapi.set_router_mac_age_limit())
        for name in names:
            r = self.nbapi.lookup("Logical_Router", name)
            self.assertEqual(r.options[ovn_const.LR_OPTIONS_MAC_AGE_LIMIT],
                             ovn_conf.get_ovn_mac_binding_age_threshold())


class TestIgnoreConnectionTimeout(BaseOvnIdlTest):
    @classmethod
    def create_connection(cls, schema):
        idl = connection.OvsdbIdl.from_server(cls.schema_map[schema], schema)
        return connection.Connection(idl, 0)

    def test_setUp_will_fail_if_this_is_broken(self):
        pass



# Copyright 2020 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_concurrency import processutils

from neutron.privileged.agent.linux import utils as priv_utils
from neutron.tests import base


NETSTAT_NETNS_OUTPUT = ("""
Active Internet connections (only servers)
Proto Recv-Q Send-Q Local Address           Foreign Address         State\
       PID/Program name
tcp        0      0 0.0.0.0:9697            0.0.0.0:*               LISTEN\
      1347/python
raw        0      0 0.0.0.0:112             0.0.0.0:*               7\
           1279/keepalived
raw        0      0 0.0.0.0:112             0.0.0.0:*               7\
           1279/keepalived
raw6       0      0 :::58                   :::*                    7\
           1349/radvd
Active UNIX domain sockets (only servers)
Proto RefCnt Flags       Type       State         I-Node   PID/Program name\
     Path
unix  2      [ ACC ]     STREAM     LISTENING     82039530 1353/python\
          /tmp/rootwrap-VKSm8a/rootwrap.sock
""")

NETSTAT_NO_NAMESPACE = ("""
Cannot open network namespace "qrouter-e6f206b2-4e8d-4597-a7e1-c3a20337e9c6":\
 No such file or directory
""")

NETSTAT_NO_LISTEN_PROCS = ("""
Active Internet connections (only servers)
Proto Recv-Q Send-Q Local Address           Foreign Address         State\
       PID/Program name
Active UNIX domain sockets (only servers)
Proto RefCnt Flags       Type       State         I-Node   PID/Program name\
     Path
""")


class FindListenPidsNamespaceTestCase(base.BaseTestCase):

    def _test_find_listen_pids_namespace_helper(self, expected,
                                                netstat_output=None):
        with mock.patch.object(processutils, 'execute') as mock_execute:
            mock_execute.return_value = (netstat_output, mock.ANY)
            observed = priv_utils._find_listen_pids_namespace(mock.ANY)
            self.assertEqual(sorted(expected), sorted(observed))

    def test_find_listen_pids_namespace_correct_output(self):
        expected = ['1347', '1279', '1349', '1353']
        self._test_find_listen_pids_namespace_helper(expected,
                                                     NETSTAT_NETNS_OUTPUT)

    def test_find_listen_pids_namespace_no_procs(self):
        self._test_find_listen_pids_namespace_helper([],
                                                     NETSTAT_NO_LISTEN_PROCS)

    def test_find_listen_pids_namespace_no_namespace(self):
        self._test_find_listen_pids_namespace_helper([], NETSTAT_NO_NAMESPACE)



# Copyright 2019 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
import sqlalchemy as sa

from neutron.db import migration


"""subnet force network id

Revision ID: c613d0b82681
Revises: 63fd95af7dcd
Create Date: 2019-08-19 11:15:14.443244

"""

# revision identifiers, used by Alembic.
revision = 'c613d0b82681'
down_revision = '63fd95af7dcd'

# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.TRAIN]


def upgrade():
    op.alter_column('subnets', 'network_id', nullable=False,
                    existing_type=sa.String(36))



# Copyright (c) 2022 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants
from neutron_lib import context
from neutron_lib.db import api as db_api
from oslo_config import cfg
from oslo_db import options as db_options
from oslo_log import log as logging

from neutron.common import config as common_config
from neutron.objects import ports as ports_obj


LOG = logging.getLogger(__name__)


def setup_conf(conf):
    common_config.register_common_config_options()
    db_group, neutron_db_opts = db_options.list_opts()[0]
    conf.register_cli_opts(neutron_db_opts, db_group)
    conf()


def main():
    """Main method for removing duplicated port binding and pb level registers.

    This script finds all ``PortBinding`` registers with the same ``port_id``.
    That happens during the live-migration process. Once finished, the inactive
    port binding register is deleted. However, it could happen that during the
    live-migration, an error occurs and this deletion is not executed. The
    related port cannot be migrated anymore.

    This script removes the inactive ``PortBinding`` referred to a port ID and
    the corresponding ``PortBindingLevel`` registers associated to this port
    ID and ``PortBinding.host``.

    This script should not be executed during a live migration process. It will
    remove the inactive port binding and will break the migration.
    """
    conf = cfg.CONF
    setup_conf(conf)
    _dry_run = conf.cli_script.dry_run
    admin_ctx = context.get_admin_context()
    with db_api.CONTEXT_WRITER.using(admin_ctx):
        dup_pbindings = ports_obj.PortBinding.get_duplicated_port_bindings(
            admin_ctx)

        # Clean duplicated port bindings that are INACTIVE and the
        # corresponding port binding level registers (if not in dry-run).
        if not _dry_run:
            for pbinding in dup_pbindings:
                port_id, host = pbinding.port_id, pbinding.host
                ports_obj.PortBinding.delete_objects(
                    admin_ctx, status=constants.INACTIVE, port_id=port_id)
                ports_obj.PortBindingLevel.delete_objects(
                    admin_ctx, port_id=port_id, host=host)

        if dup_pbindings:
            port_ids = [pbinding.port_id for pbinding in dup_pbindings]
            action = 'can be' if _dry_run else 'have been'
            LOG.info('The following duplicated PortBinding registers with '
                     'status=INACTIVE %s removed, port_ids: %s',
                     action, port_ids)
        else:
            LOG.info('No duplicated PortBinding registers has been found.')



# Copyright (c) 2017 Fujitsu Limited
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib import constants
from neutron_lib.plugins.ml2 import ovs_constants as ovs_consts
from oslo_config import cfg
from oslo_utils import uuidutils

from neutron.objects.logapi import logging_resource as log_object
from neutron.services.logapi.common import exceptions as log_exc
from neutron.services.logapi.drivers.openvswitch \
    import ovs_firewall_log as ovsfw_log
from neutron.services.logapi.rpc import agent as agent_rpc
from neutron.tests import base
from neutron.tests import tools

COOKIE_ID = uuidutils.generate_uuid()
PORT_ID = uuidutils.generate_uuid()
PROJECT_ID = uuidutils.generate_uuid()
ACTION = tools.get_random_security_event()
LOG_ID = uuidutils.generate_uuid()
SG_ID = uuidutils.generate_uuid()
REMOTE_SG_ID = uuidutils.generate_uuid()

FakeSGLogInfo = [
    {
        'id': LOG_ID,
        'ports_log': [{'port_id': PORT_ID,
                       'security_group_rules': [
                           {'ethertype': constants.IPv4,
                            'protocol': constants.PROTO_NAME_TCP,
                            'direction': constants.INGRESS_DIRECTION,
                            'port_range_min': 123,
                            'port_range_max': 123,
                            'security_group_id': SG_ID},
                           {'ethertype': constants.IPv4,
                            'protocol': constants.PROTO_NAME_UDP,
                            'direction': constants.EGRESS_DIRECTION,
                            'security_group_id': SG_ID},
                           {'ethertype': constants.IPv6,
                            'protocol': constants.PROTO_NAME_TCP,
                            'remote_group_id': REMOTE_SG_ID,
                            'direction': constants.EGRESS_DIRECTION,
                            'security_group_id': SG_ID}
                       ]}],
        'event': 'ALL',
        'project_id': PROJECT_ID,
    }
]


def set_log_driver_config(ctrl_rate_limit, ctrl_burst_limit):
    cfg.CONF.set_override('rate_limit', ctrl_rate_limit, group='network_log')
    cfg.CONF.set_override('burst_limit', ctrl_burst_limit, group='network_log')


class TestCookie(base.BaseTestCase):
    def setUp(self):
        super(TestCookie, self).setUp()
        self.cookie = ovsfw_log.Cookie(COOKIE_ID, PORT_ID, ACTION, PROJECT_ID)
        self.cookie.log_object_refs = set([LOG_ID])

    def test_add_log_object_refs(self):
        new_log_id = uuidutils.generate_uuid()
        expected = set([LOG_ID, new_log_id])
        self.cookie.add_log_obj_ref(new_log_id)
        self.assertEqual(expected, self.cookie.log_object_refs)

    def test_removed_log_object_ref(self):
        expected = set()
        self.cookie.remove_log_obj_ref(LOG_ID)
        self.assertEqual(expected, self.cookie.log_object_refs)

    def test_is_empty(self):
        self.cookie.remove_log_obj_ref(LOG_ID)
        result = self.cookie.is_empty
        self.assertTrue(result)


class FakeOVSPort(object):
    def __init__(self, name, port, mac):
        self.port_name = name
        self.ofport = port
        self.vif_mac = mac


class TestOVSFirewallLoggingDriver(base.BaseTestCase):
    def setUp(self):
        super(TestOVSFirewallLoggingDriver, self).setUp()
        mock_int_br = mock.Mock()
        mock_int_br.br.dump_flows.return_value = []
        self._mock_initialize_bridge = mock.patch.object(
            ovsfw_log.OVSFirewallLoggingDriver, 'initialize_bridge',
            return_value=mock_int_br)
        self.mock_initialize_bridge = self._mock_initialize_bridge.start()
        self.log_driver = ovsfw_log.OVSFirewallLoggingDriver(mock.Mock())
        resource_rpc_mock = mock.patch.object(
            agent_rpc, 'LoggingApiStub', autospec=True).start()
        self.log_driver.start_logapp = mock.Mock()
        self.log_driver.initialize(resource_rpc_mock)
        self.log_driver.SUPPORTED_LOGGING_TYPES = ['security_group']
        self.mock_bridge = self.log_driver.int_br
        self.mock_bridge.reset_mock()
        self.fake_ovs_port = FakeOVSPort('port', 1, '00:00:00:00:00:00')
        self.mock_bridge.br.get_vif_port_by_id.return_value = \
            self.fake_ovs_port
        log_data = {
            'context': None,
            'name': 'test1',
            'id': LOG_ID,
            'project_id': PROJECT_ID,
            'event': 'ALL',
            'resource_type': 'security_group'
        }
        self.log_resource = log_object.Log(**log_data)

    @property
    def port_ofport(self):
        return self.mock_bridge.br.get_vif_port_by_id.return_value.ofport

    @property
    def port_mac(self):
        return self.mock_bridge.br.get_vif_port_by_id.return_value.vif_mac

    def test_initialize_bridge(self):
        self._mock_initialize_bridge.stop()
        br = self.log_driver.initialize_bridge(self.mock_bridge)
        self.assertEqual(self.mock_bridge.deferred.return_value, br)

    def test_set_controller_rate_limit(self):
        self._mock_initialize_bridge.stop()
        set_log_driver_config(100, 25)
        self.log_driver.initialize_bridge(self.mock_bridge)
        expected_calls = [mock.call.set_controller_rate_limit(100),
                          mock.call.set_controller_burst_limit(25)]
        self.mock_bridge.assert_has_calls(expected_calls)

    def test_generate_cookie(self):
        cookie_id = self.log_driver.generate_cookie(
            PORT_ID, ACTION, LOG_ID, PROJECT_ID)
        cookie = self.log_driver._get_cookie_by_id(cookie_id)
        self.assertIn(cookie, self.log_driver.cookies_table)

    def test__get_cookie_by_id_not_found(self):
        cookie_id = uuidutils.generate_uuid()
        cookie = ovsfw_log.Cookie(cookie_id=uuidutils.generate_uuid(),
                                  port=PORT_ID, action=ACTION,
                                  project=PROJECT_ID)
        self.log_driver.cookies_table = set([cookie])
        self.assertRaises(log_exc.CookieNotFound,
                          self.log_driver._get_cookie_by_id,
                          cookie_id)

    def test_start_log_with_update_or_create_log_event(self):
        context = mock.Mock()
        log_data = {'log_resources': [self.log_resource]}
        self.log_driver.resource_rpc.get_sg_log_info_for_log_resources.\
            return_value = FakeSGLogInfo
        self.log_driver.start_logging(context, **log_data)
        accept_cookie = self.log_driver._get_cookie(PORT_ID, 'ACCEPT')
        drop_cookie = self.log_driver._get_cookie(PORT_ID, 'DROP')
        conj_id = self.log_driver.conj_id_map.get_conj_id(
            SG_ID, REMOTE_SG_ID, constants.EGRESS_DIRECTION, constants.IPv6)
        add_rules = [
            # log ingress tcp port=123
            mock.call(
                actions='controller',
                cookie=accept_cookie.id,
                reg5=self.port_ofport,
                dl_type="0x{:04x}".format(constants.ETHERTYPE_IP),
                nw_proto=constants.PROTO_NUM_TCP,
                priority=77,
                table=ovs_consts.ACCEPTED_INGRESS_TRAFFIC_TABLE,
                tcp_dst='0x007b'),
            # log egress tcp6
            mock.call(
                actions='resubmit(,%d),controller' % (
                    ovs_consts.ACCEPTED_EGRESS_TRAFFIC_NORMAL_TABLE),
                cookie=accept_cookie.id,
                reg5=self.port_ofport,
                dl_type="0x{:04x}".format(constants.ETHERTYPE_IPV6),
                priority=70,
                reg7=conj_id + 1,
                table=ovs_consts.ACCEPTED_EGRESS_TRAFFIC_TABLE),
            # log egress udp
            mock.call(
                actions='resubmit(,%d),controller' % (
                    ovs_consts.ACCEPTED_EGRESS_TRAFFIC_NORMAL_TABLE),
                cookie=accept_cookie.id,
                reg5=self.port_ofport,
                dl_type="0x{:04x}".format(constants.ETHERTYPE_IP),
                nw_proto=constants.PROTO_NUM_UDP,
                priority=77,
                table=ovs_consts.ACCEPTED_EGRESS_TRAFFIC_TABLE,
            ),
            # log drop
            mock.call(
                actions='controller',
                cookie=drop_cookie.id,
                priority=53,
                reg5=self.port_ofport,
                table=ovs_consts.DROPPED_TRAFFIC_TABLE,
            )

        ]
        self.mock_bridge.br.add_flow.assert_has_calls(
            add_rules, any_order=True)

    def test_stop_log_with_delete_log_event(self):
        context = mock.Mock()
        log_data = {'log_resources': [self.log_resource]}
        self.log_driver.resource_rpc.get_sg_log_info_for_log_resources.\
            return_value = FakeSGLogInfo
        self.log_driver.start_logging(context, **log_data)
        accept_cookie = self.log_driver._get_cookie(PORT_ID, 'ACCEPT')
        drop_cookie = self.log_driver._get_cookie(PORT_ID, 'DROP')
        self.mock_bridge.reset_mock()
        self.log_driver.stop_logging(context, **log_data)

        delete_rules = [
            # delete drop flow
            mock.call(
                table=ovs_consts.DROPPED_TRAFFIC_TABLE,
                cookie=drop_cookie.id
            ),
            # delete accept flows
            mock.call(
                table=ovs_consts.ACCEPTED_EGRESS_TRAFFIC_TABLE,
                cookie=accept_cookie.id
            ),
            mock.call(
                table=ovs_consts.ACCEPTED_INGRESS_TRAFFIC_TABLE,
                cookie=accept_cookie.id
            )
        ]

        self.mock_bridge.br.delete_flows.assert_has_calls(
            delete_rules, any_order=True)

    def test_start_log_with_add_port_event(self):
        context = mock.Mock()
        log_data = {'port_id': PORT_ID}
        self.log_driver.resource_rpc.get_sg_log_info_for_port.return_value = \
            [
                {
                    'id': uuidutils.generate_uuid(),
                    'ports_log': [{'port_id': PORT_ID,
                                   'security_group_rules': [
                                       {'ethertype': constants.IPv4,
                                        'protocol': constants.PROTO_NAME_TCP,
                                        'direction':
                                            constants.INGRESS_DIRECTION,
                                        'port_range_min': 123,
                                        'port_range_max': 123,
                                        'security_group_id': 456}]}],
                    'event': 'ACCEPT',
                    'project_id': PROJECT_ID,
                }
            ]
        self.log_driver.start_logging(context, **log_data)
        accept_cookie = self.log_driver._get_cookie(PORT_ID, 'ACCEPT')
        add_rules = [
            # log ingress tcp port=123
            mock.call(
                actions='controller',
                cookie=accept_cookie.id,
                reg5=self.port_ofport,
                dl_type="0x{:04x}".format(constants.ETHERTYPE_IP),
                nw_proto=constants.PROTO_NUM_TCP,
                priority=77,
                table=ovs_consts.ACCEPTED_INGRESS_TRAFFIC_TABLE,
                tcp_dst='0x007b')
        ]
        self.mock_bridge.br.add_flow.assert_has_calls(
            add_rules, any_order=True)

    def test_stop_log_with_delete_port_event(self):

        context = mock.Mock()
        log_data = {'port_id': PORT_ID}
        # add port
        self.log_driver.resource_rpc.get_sg_log_info_for_port.return_value = \
            FakeSGLogInfo
        self.log_driver.start_logging(context, **log_data)
        accept_cookie = self.log_driver._get_cookie(PORT_ID, 'ACCEPT')
        drop_cookie = self.log_driver._get_cookie(PORT_ID, 'DROP')
        self.mock_bridge.reset_mock()
        # delete port
        self.log_driver.stop_logging(
            context, port_id=PORT_ID)

        delete_rules = [
            # delete accept flows
            mock.call(
                table=ovs_consts.ACCEPTED_INGRESS_TRAFFIC_TABLE,
                cookie=accept_cookie.id
            ),
            mock.call(
                table=ovs_consts.ACCEPTED_EGRESS_TRAFFIC_TABLE,
                cookie=accept_cookie.id
            ),
            # delete drop flow
            mock.call(
                table=ovs_consts.DROPPED_TRAFFIC_TABLE,
                cookie=drop_cookie.id
            ),
        ]
        self.mock_bridge.br.delete_flows.assert_has_calls(
            delete_rules, any_order=True)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import abc
import copy
import random

from oslo_log import log

from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf


LOG = log.getLogger(__name__)

OVN_SCHEDULER_CHANCE = 'chance'
OVN_SCHEDULER_LEAST_LOADED = 'leastloaded'


class OVNGatewayScheduler(object, metaclass=abc.ABCMeta):

    def __init__(self):
        pass

    @abc.abstractmethod
    def select(self, nb_idl, gateway_name, candidates=None):
        """Schedule the gateway port of a router to an OVN chassis.

        Schedule the gateway router port only if it is not already
        scheduled.
        """

    def filter_existing_chassis(self, nb_idl, gw_chassis,
                                physnet, chassis_physnets,
                                existing_chassis, az_hints, chassis_with_azs):
        chassis_list = copy.copy(existing_chassis)
        for chassis_name in existing_chassis:
            if utils.is_gateway_chassis_invalid(chassis_name, gw_chassis,
                                                physnet, chassis_physnets,
                                                az_hints, chassis_with_azs):
                LOG.debug("Chassis %(chassis)s is invalid for scheduling "
                          "router in physnet: %(physnet)s.",
                          {'chassis': chassis_name,
                           'physnet': physnet})
                chassis_list.remove(chassis_name)
        return chassis_list

    def _schedule_gateway(self, nb_idl, gateway_name, candidates,
                          existing_chassis):
        existing_chassis = existing_chassis or []
        candidates = candidates or []
        candidates = list(set(candidates) - set(existing_chassis))
        # If no candidates, or gateway scheduled on MAX_GATEWAY_CHASSIS nodes
        # or all candidates in existing_chassis, return existing_chassis.
        # Otherwise, if more candidates present, then schedule them.
        if existing_chassis:
            if not candidates or (
                    len(existing_chassis) == ovn_const.MAX_GW_CHASSIS):
                return existing_chassis
        if not candidates:
            LOG.warning('Gateway %s was not scheduled on any chassis, no '
                        'candidates are available', gateway_name)
            return [ovn_const.OVN_GATEWAY_INVALID_CHASSIS]
        chassis_count = ovn_const.MAX_GW_CHASSIS - len(existing_chassis)
        # The actual binding of the gateway to a chassis via the options
        # column or gateway_chassis column in the OVN_Northbound is done
        # by the caller
        chassis = self._select_gateway_chassis(
            nb_idl, candidates)[:chassis_count]
        # priority of existing chassis is higher than candidates
        chassis = existing_chassis + chassis

        LOG.debug("Gateway %s scheduled on chassis %s",
                  gateway_name, chassis)
        return chassis

    @abc.abstractmethod
    def _select_gateway_chassis(self, nb_idl, candidates):
        """Choose a chassis from candidates based on a specific policy."""


class OVNGatewayChanceScheduler(OVNGatewayScheduler):
    """Randomly select an chassis for a gateway port of a router"""

    def select(self, nb_idl, gateway_name, candidates=None,
               existing_chassis=None):
        return self._schedule_gateway(nb_idl, gateway_name,
                                      candidates, existing_chassis)

    def _select_gateway_chassis(self, nb_idl, candidates):
        candidates = copy.deepcopy(candidates)
        random.shuffle(candidates)
        return candidates


class OVNGatewayLeastLoadedScheduler(OVNGatewayScheduler):
    """Select the least loaded chassis for a gateway port of a router"""

    def select(self, nb_idl, gateway_name, candidates=None,
               existing_chassis=None):
        return self._schedule_gateway(nb_idl, gateway_name,
                                      candidates, existing_chassis)

    @staticmethod
    def _get_chassis_load_by_prios(chassis_info):
        """Retrieve the amount of ports by priorities hosted in the chassis.

        @param   chassis_info: list of (port, prio) hosted by this chassis
        @type    chassis_info: []
        @return: A list of (prio, number_of_ports) tuples.
        """
        chassis_load = {}
        for lrp, prio in chassis_info:
            chassis_load[prio] = chassis_load.get(prio, 0) + 1
        return chassis_load.items()

    @staticmethod
    def _get_chassis_load(chassis):
        chassis_ports_prios = chassis[1]
        return sorted(
            OVNGatewayLeastLoadedScheduler._get_chassis_load_by_prios(
                chassis_ports_prios), reverse=True)

    def _select_gateway_chassis(self, nb_idl, candidates):
        chassis_bindings = nb_idl.get_all_chassis_gateway_bindings(candidates)
        return [chassis for chassis, load in
                sorted(chassis_bindings.items(),
                       key=OVNGatewayLeastLoadedScheduler._get_chassis_load)]


OVN_SCHEDULER_STR_TO_CLASS = {
    OVN_SCHEDULER_CHANCE: OVNGatewayChanceScheduler,
    OVN_SCHEDULER_LEAST_LOADED: OVNGatewayLeastLoadedScheduler}


def get_scheduler():
    return OVN_SCHEDULER_STR_TO_CLASS[ovn_conf.get_ovn_l3_scheduler()]()



# Copyright 2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from neutron_lib.api.definitions import portbindings
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib import constants
from neutron_lib import context
from neutron_lib.plugins.ml2 import ovs_constants as agent_consts
from neutron_lib.services.trunk import constants as trunk_consts
from oslo_config import cfg

from neutron.services.trunk.drivers.openvswitch import driver
from neutron.tests import base

GEN_TRUNK_BR_NAME_PATCH = (
    'neutron.services.trunk.drivers.openvswitch.utils.gen_trunk_br_name')


class OVSDriverTestCase(base.BaseTestCase):

    def test_driver_creation(self):
        ovs_driver = driver.OVSDriver.create()
        self.assertFalse(ovs_driver.is_loaded)
        self.assertEqual(driver.NAME, ovs_driver.name)
        self.assertEqual(driver.SUPPORTED_INTERFACES, ovs_driver.interfaces)
        self.assertEqual(driver.SUPPORTED_SEGMENTATION_TYPES,
                         ovs_driver.segmentation_types)
        self.assertEqual(constants.AGENT_TYPE_OVS, ovs_driver.agent_type)
        self.assertFalse(ovs_driver.can_trunk_bound_port)
        self.assertTrue(
            ovs_driver.is_agent_compatible(constants.AGENT_TYPE_OVS))
        self.assertTrue(
            ovs_driver.is_interface_compatible(driver.SUPPORTED_INTERFACES[0]))

    def test_driver_is_loaded(self):
        cfg.CONF.set_override('mechanism_drivers',
                              'openvswitch', group='ml2')
        ovs_driver = driver.OVSDriver.create()
        self.assertTrue(ovs_driver.is_loaded)

    def test_driver_is_not_loaded(self):
        cfg.CONF.set_override('core_plugin', 'my_foo_plugin')
        ovs_driver = driver.OVSDriver.create()
        self.assertFalse(ovs_driver.is_loaded)

    @mock.patch(GEN_TRUNK_BR_NAME_PATCH)
    def test_vif_details_bridge_name_handler_registration(self,
                                                          mock_gen_br_name):
        driver.register()
        mock_gen_br_name.return_value = 'fake-trunk-br-name'
        test_trigger = mock.Mock()
        registry.publish(agent_consts.OVS_BRIDGE_NAME, events.BEFORE_READ,
                         test_trigger,
                         payload=events.EventPayload(
                             None, metadata={
                                 'port': {
                                     'trunk_details': {
                                         'trunk_id': 'foo'
                                     }
                                 }
                             }))
        test_trigger.assert_called_once_with('fake-trunk-br-name')

    def test__update_subport_binding(self):
        ovs_driver = driver.OVSDriver.create()
        core_plugin = mock.Mock()
        core_plugin.get_port.return_value = {portbindings.HOST_ID: 'host_id'}
        ovs_driver._core_plugin = core_plugin
        ctx = context.get_admin_context()
        mock_subports = [mock.Mock(port_id='sport_id1'),
                         mock.Mock(port_id='sport_id2')]
        mock_trunk_obj = mock.Mock(sub_ports=mock_subports, port_id='port_id',
                                   id='trunk_id')
        with mock.patch.object(ovs_driver, '_get_trunk') as mock_get_trunk:
            mock_get_trunk.return_value = mock_trunk_obj
            updated_port = {
                'port': {portbindings.HOST_ID: 'host_id',
                         'device_owner': trunk_consts.TRUNK_SUBPORT_OWNER}}
            ovs_driver._update_subport_binding(ctx, 'trunk_id')
            core_plugin.update_port.assert_has_calls(
                [mock.call(ctx, 'sport_id1', updated_port),
                 mock.call(ctx, 'sport_id2', updated_port)], any_order=True)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib import constants as const
from ovsdbapp.backend.ovs_idl import idlutils

from neutron.common.ovn import acl as ovn_acl
from neutron.common.ovn import utils as ovn_utils
from neutron.conf.agent import securitygroups_rpc
from neutron.tests import base
from neutron.tests.unit import fake_resources as fakes


class TestACLs(base.BaseTestCase):

    def setUp(self):
        super(TestACLs, self).setUp()
        self.driver = mock.Mock()
        self.driver.nb_ovn = fakes.FakeOvsdbNbOvnIdl()
        self.plugin = fakes.FakePlugin()
        self.admin_context = mock.Mock()
        self.fake_port = fakes.FakePort.create_one_port({
            'id': 'fake_port_id1',
            'network_id': 'network_id1',
            'fixed_ips': [{'subnet_id': 'subnet_id1',
                           'ip_address': '1.1.1.1'}],
        }).info()
        self.fake_subnet = fakes.FakeSubnet.create_one_subnet({
            'id': 'subnet_id1',
            'ip_version': 4,
            'cidr': '1.1.1.0/24',
        }).info()
        mock_row_by_value = mock.patch.object(idlutils, 'row_by_value')
        mock_row_by_value.start()
        self.addCleanup(mock_row_by_value.stop)
        mock_acl_columns_severity = mock.patch.object(
            ovn_acl, '_acl_columns_name_severity_supported', return_value=True)
        mock_acl_columns_severity.start()
        self.addCleanup(mock_acl_columns_severity.stop)
        securitygroups_rpc.register_securitygroups_opts()

    def test_drop_all_ip_traffic_for_port(self):
        acls = ovn_acl.drop_all_ip_traffic_for_port(self.fake_port)
        acl_to_lport = {'action': 'drop', 'direction': 'to-lport',
                        'external_ids': {'neutron:lport':
                                         self.fake_port['id']},
                        'log': False, 'name': [], 'severity': [],
                        'lport': self.fake_port['id'],
                        'lswitch': 'neutron-network_id1',
                        'match': 'outport == "fake_port_id1" && ip',
                        'priority': 1001}
        acl_from_lport = {'action': 'drop', 'direction': 'from-lport',
                          'external_ids': {'neutron:lport':
                                           self.fake_port['id']},
                          'log': False, 'name': [], 'severity': [],
                          'lport': self.fake_port['id'],
                          'lswitch': 'neutron-network_id1',
                          'match': 'inport == "fake_port_id1" && ip',
                          'priority': 1001}
        for acl in acls:
            if 'to-lport' in acl.values():
                self.assertEqual(acl_to_lport, acl)
            if 'from-lport' in acl.values():
                self.assertEqual(acl_from_lport, acl)

    def test_acl_protocol_and_ports_for_tcp_udp_and_sctp_number(self):
        sg_rule = {'port_range_min': None,
                   'port_range_max': None}

        sg_rule['protocol'] = str(const.PROTO_NUM_TCP)
        match = ovn_acl.acl_protocol_and_ports(sg_rule, None)
        self.assertEqual(' && tcp', match)

        sg_rule['protocol'] = str(const.PROTO_NUM_UDP)
        match = ovn_acl.acl_protocol_and_ports(sg_rule, None)
        self.assertEqual(' && udp', match)

        sg_rule['protocol'] = str(const.PROTO_NUM_SCTP)
        match = ovn_acl.acl_protocol_and_ports(sg_rule, None)
        self.assertEqual(' && sctp', match)

    def test_acl_protocol_and_ports_for_tcp_udp_and_sctp_number_one(self):
        sg_rule = {'port_range_min': 22,
                   'port_range_max': 22}

        sg_rule['protocol'] = str(const.PROTO_NUM_TCP)
        match = ovn_acl.acl_protocol_and_ports(sg_rule, None)
        self.assertEqual(' && tcp && tcp.dst == 22', match)

        sg_rule['protocol'] = str(const.PROTO_NUM_UDP)
        match = ovn_acl.acl_protocol_and_ports(sg_rule, None)
        self.assertEqual(' && udp && udp.dst == 22', match)

        sg_rule['protocol'] = str(const.PROTO_NUM_SCTP)
        match = ovn_acl.acl_protocol_and_ports(sg_rule, None)
        self.assertEqual(' && sctp && sctp.dst == 22', match)

    def test_acl_protocol_and_ports_for_tcp_udp_and_sctp_number_range(self):
        sg_rule = {'port_range_min': 21,
                   'port_range_max': 23}

        sg_rule['protocol'] = str(const.PROTO_NUM_TCP)
        match = ovn_acl.acl_protocol_and_ports(sg_rule, None)
        self.assertEqual(' && tcp && tcp.dst >= 21 && tcp.dst <= 23', match)

        sg_rule['protocol'] = str(const.PROTO_NUM_UDP)
        match = ovn_acl.acl_protocol_and_ports(sg_rule, None)
        self.assertEqual(' && udp && udp.dst >= 21 && udp.dst <= 23', match)

        sg_rule['protocol'] = str(const.PROTO_NUM_SCTP)
        match = ovn_acl.acl_protocol_and_ports(sg_rule, None)
        self.assertEqual(' && sctp && sctp.dst >= 21 && sctp.dst <= 23', match)

    def test_acl_protocol_and_ports_for_ipv6_icmp_protocol(self):
        sg_rule = {'port_range_min': None,
                   'port_range_max': None}
        icmp = 'icmp6'
        expected_match = ' && icmp6'

        sg_rule['protocol'] = const.PROTO_NAME_ICMP
        match = ovn_acl.acl_protocol_and_ports(sg_rule, icmp)
        self.assertEqual(expected_match, match)

        sg_rule['protocol'] = str(const.PROTO_NUM_ICMP)
        match = ovn_acl.acl_protocol_and_ports(sg_rule, icmp)
        self.assertEqual(expected_match, match)

        sg_rule['protocol'] = const.PROTO_NAME_IPV6_ICMP
        match = ovn_acl.acl_protocol_and_ports(sg_rule, icmp)
        self.assertEqual(expected_match, match)

        sg_rule['protocol'] = const.PROTO_NAME_IPV6_ICMP_LEGACY
        match = ovn_acl.acl_protocol_and_ports(sg_rule, icmp)
        self.assertEqual(expected_match, match)

        sg_rule['protocol'] = str(const.PROTO_NUM_IPV6_ICMP)
        match = ovn_acl.acl_protocol_and_ports(sg_rule, icmp)
        self.assertEqual(expected_match, match)

    def test_acl_protocol_and_ports_for_icmp4_and_icmp6_port_range(self):
        match_list = [
            (None, None, ' && icmp4'),
            (0, None, ' && icmp4 && icmp4.type == 0'),
            (0, 0, ' && icmp4 && icmp4.type == 0 && icmp4.code == 0'),
            (0, 5, ' && icmp4 && icmp4.type == 0 && icmp4.code == 5')]
        v6_match_list = [
            (None, None, ' && icmp6'),
            (133, None, ' && icmp6 && icmp6.type == 133'),
            (1, 1, ' && icmp6 && icmp6.type == 1 && icmp6.code == 1'),
            (138, 1, ' && icmp6 && icmp6.type == 138 && icmp6.code == 1')]

        sg_rule = {'protocol': const.PROTO_NAME_ICMP}
        icmp = 'icmp4'
        for pmin, pmax, expected_match in match_list:
            sg_rule['port_range_min'] = pmin
            sg_rule['port_range_max'] = pmax
            match = ovn_acl.acl_protocol_and_ports(sg_rule, icmp)
            self.assertEqual(expected_match, match)

        sg_rule = {'protocol': const.PROTO_NAME_IPV6_ICMP}
        icmp = 'icmp6'
        for pmin, pmax, expected_match in v6_match_list:
            sg_rule['port_range_min'] = pmin
            sg_rule['port_range_max'] = pmax
            match = ovn_acl.acl_protocol_and_ports(sg_rule, icmp)
            self.assertEqual(expected_match, match)

    def test_acl_protocol_and_ports_protocol_not_supported(self):
        sg_rule = {'port_range_min': None,
                   'port_range_max': None}
        sg_rule['protocol'] = '1234567'
        self.assertRaises(ovn_acl.ProtocolNotSupported,
                          ovn_acl.acl_protocol_and_ports, sg_rule, None)

    def test_acl_protocol_and_ports_protocol_range(self):
        sg_rule = {'port_range_min': None,
                   'port_range_max': None}

        # For more common protocols such as TCP, UDP and ICMP, we
        # prefer to use the protocol name in the match string instead of
        # the protocol number (e.g: the word "tcp" instead of "ip.proto
        # == 6"). This improves the readability/debbugability when
        # troubleshooting the ACLs
        skip_protos = (const.PROTO_NUM_TCP, const.PROTO_NUM_UDP,
                       const.PROTO_NUM_SCTP, const.PROTO_NUM_ICMP,
                       const.PROTO_NUM_IPV6_ICMP)

        for proto in range(256):
            if proto in skip_protos:
                continue
            sg_rule['protocol'] = str(proto)
            match = ovn_acl.acl_protocol_and_ports(sg_rule, None)
            self.assertEqual(' && ip.proto == %s' % proto, match)

    def test_acl_protocol_and_ports_name_to_number(self):
        sg_rule = {'port_range_min': None,
                   'port_range_max': None}

        sg_rule['protocol'] = str(const.PROTO_NAME_OSPF)
        match = ovn_acl.acl_protocol_and_ports(sg_rule, None)
        self.assertEqual(' && ip.proto == 89', match)

    def test_acl_direction(self):
        sg_rule = fakes.FakeSecurityGroupRule.create_one_security_group_rule({
            'direction': 'ingress'
        }).info()

        match = ovn_acl.acl_direction(sg_rule, self.fake_port)
        self.assertEqual('outport == "' + self.fake_port['id'] + '"', match)

        sg_rule['direction'] = 'egress'
        match = ovn_acl.acl_direction(sg_rule, self.fake_port)
        self.assertEqual('inport == "' + self.fake_port['id'] + '"', match)

    def test_acl_ethertype(self):
        sg_rule = fakes.FakeSecurityGroupRule.create_one_security_group_rule({
            'ethertype': 'IPv4'
        }).info()

        match, ip_version, icmp = ovn_acl.acl_ethertype(sg_rule)
        self.assertEqual(' && ip4', match)
        self.assertEqual('ip4', ip_version)
        self.assertEqual('icmp4', icmp)

        sg_rule['ethertype'] = 'IPv6'
        match, ip_version, icmp = ovn_acl.acl_ethertype(sg_rule)
        self.assertEqual(' && ip6', match)
        self.assertEqual('ip6', ip_version)
        self.assertEqual('icmp6', icmp)

        sg_rule['ethertype'] = 'IPv10'
        match, ip_version, icmp = ovn_acl.acl_ethertype(sg_rule)
        self.assertEqual('', match)
        self.assertIsNone(ip_version)
        self.assertIsNone(icmp)

    def test_acl_remote_ip_prefix(self):
        sg_rule = fakes.FakeSecurityGroupRule.create_one_security_group_rule({
            'direction': 'ingress',
            'remote_ip_prefix': None
        }).info()
        ip_version = 'ip4'
        remote_ip_prefix = '10.10.0.0/24'
        normalized_cidr = '10.10.0.0/24'

        match = ovn_acl.acl_remote_ip_prefix(sg_rule, ip_version)
        self.assertEqual('', match)

        sg_rule['remote_ip_prefix'] = remote_ip_prefix
        sg_rule['normalized_cidr'] = normalized_cidr
        match = ovn_acl.acl_remote_ip_prefix(sg_rule, ip_version)
        expected_match = ' && %s.src == %s' % (ip_version, remote_ip_prefix)
        self.assertEqual(expected_match, match)

        sg_rule['direction'] = 'egress'
        match = ovn_acl.acl_remote_ip_prefix(sg_rule, ip_version)
        expected_match = ' && %s.dst == %s' % (ip_version, remote_ip_prefix)
        self.assertEqual(expected_match, match)

    def test_acl_remote_ip_prefix_not_normalized(self):
        normalized_ip_prefix = '10.10.10.128/26'
        ip_version = 'ip4'
        sg_rule = fakes.FakeSecurityGroupRule.create_one_security_group_rule({
            'direction': 'ingress',
            'remote_ip_prefix': '10.10.10.175/26',
            'normalized_cidr': normalized_ip_prefix
        }).info()

        match = ovn_acl.acl_remote_ip_prefix(sg_rule, ip_version)
        expected_match = ' && %s.src == %s' % (ip_version,
                                               normalized_ip_prefix)
        self.assertEqual(expected_match, match)

    def test_acl_remote_group_id(self):
        sg_rule = fakes.FakeSecurityGroupRule.create_one_security_group_rule({
            'direction': 'ingress',
            'remote_group_id': None
        }).info()
        ip_version = 'ip4'
        sg_id = sg_rule['security_group_id']

        pg_name = ovn_utils.ovn_pg_addrset_name(sg_id, ip_version)

        match = ovn_acl.acl_remote_group_id(sg_rule, ip_version)
        self.assertEqual('', match)

        sg_rule['remote_group_id'] = sg_id
        match = ovn_acl.acl_remote_group_id(sg_rule, ip_version)
        self.assertEqual(' && ip4.src == $' + pg_name, match)

        sg_rule['direction'] = 'egress'
        match = ovn_acl.acl_remote_group_id(sg_rule, ip_version)
        self.assertEqual(' && ip4.dst == $' + pg_name, match)



# Copyright 2016 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import os
import shutil

import fixtures
import psutil
import tenacity

from neutron.agent.linux import utils


class DaemonProcessFixture(fixtures.Fixture):
    def __init__(self, temp_dir):
        super(DaemonProcessFixture, self).__init__()
        self.temp_dir = temp_dir

    def _get_pid_from_pidfile(self, pidfile):
        with open(os.path.join(self.temp_dir, pidfile), 'r') as pidfile_f:
            pid = pidfile_f.read().strip()
            try:
                return int(pid)
            except ValueError:
                raise RuntimeError(
                    "Pidfile %(pidfile)s contains %(pid)s that "
                    "is not a pid" % {'pidfile': pidfile, 'pid': pid}
                )


class OvnNorthd(DaemonProcessFixture):

    def __init__(self, temp_dir, ovn_nb_db, ovn_sb_db, protocol='unix',
                 debug=True):
        super(OvnNorthd, self).__init__(temp_dir)
        self.ovn_nb_db = ovn_nb_db
        self.ovn_sb_db = ovn_sb_db
        self.protocol = protocol
        self.unixctl_path = os.path.join(self.temp_dir, 'ovn_northd.ctl')
        self.log_file_path = os.path.join(self.temp_dir, 'ovn_northd.log')
        self.debug = debug
        if self.protocol == 'ssl':
            self.private_key = os.path.join(self.temp_dir, 'ovn-privkey.pem')
            self.certificate = os.path.join(self.temp_dir, 'ovn-cert.pem')
            self.ca_cert = os.path.join(self.temp_dir, 'controllerca',
                                        'cacert.pem')

    def _setUp(self):
        self.addCleanup(self.stop)
        self.start()

    def start(self):
        # start the ovn-northd
        ovn_northd_cmd = [
            shutil.which('ovn-northd'), '-vconsole:off',
            '--detach',
            '--ovnnb-db=%s' % self.ovn_nb_db,
            '--ovnsb-db=%s' % self.ovn_sb_db,
            '--no-chdir',
            '--unixctl=%s' % self.unixctl_path,
            '--log-file=%s' % (self.log_file_path)]
        if self.protocol == 'ssl':
            ovn_northd_cmd.append('--private-key=%s' % self.private_key)
            ovn_northd_cmd.append('--certificate=%s' % self.certificate)
            ovn_northd_cmd.append('--ca-cert=%s' % self.ca_cert)
        if self.debug:
            ovn_northd_cmd.append('--verbose')
        obj, _ = utils.create_process(ovn_northd_cmd)
        obj.communicate()

    def stop(self):
        try:
            if os.path.exists(self.unixctl_path):
                stop_cmd = ['ovs-appctl', '-t', self.unixctl_path, 'exit']
                utils.execute(stop_cmd)
        except Exception:
            pass


class OvsdbServer(DaemonProcessFixture):

    def __init__(self, temp_dir, ovs_dir, ovn_nb_db=True, ovn_sb_db=False,
                 protocol='unix', debug=True):
        super(OvsdbServer, self).__init__(temp_dir)
        self.ovs_dir = ovs_dir
        self.ovn_nb_db = ovn_nb_db
        self.ovn_sb_db = ovn_sb_db
        # The value of the protocol must be unix or tcp or ssl
        self.protocol = protocol
        self.ovsdb_server_processes = []
        self.private_key = os.path.join(self.temp_dir, 'ovn-privkey.pem')
        self.certificate = os.path.join(self.temp_dir, 'ovn-cert.pem')
        self.ca_cert = os.path.join(self.temp_dir, 'controllerca',
                                    'cacert.pem')
        self.debug = debug

    def _setUp(self):
        if self.ovn_nb_db:
            self.ovsdb_server_processes.append(
                {'db_path': os.path.join(self.temp_dir, 'ovn_nb.db'),
                 'schema_path': os.path.join(self.ovs_dir, 'ovn-nb.ovsschema'),
                 'remote_path': os.path.join(self.temp_dir, 'ovnnb_db.sock'),
                 'protocol': self.protocol,
                 'remote_ip': '127.0.0.1',
                 'remote_port': '0',
                 'pidfile': 'ovn-nb.pid',
                 'unixctl_path': os.path.join(self.temp_dir, 'ovnnb_db.ctl'),
                 'log_file_path': os.path.join(self.temp_dir, 'ovn_nb.log'),
                 'db_type': 'nb',
                 'connection': 'db:OVN_Northbound,NB_Global,connections',
                 'ctl_cmd': 'ovn-nbctl'})

        if self.ovn_sb_db:
            self.ovsdb_server_processes.append(
                {'db_path': os.path.join(self.temp_dir, 'ovn_sb.db'),
                 'schema_path': os.path.join(self.ovs_dir, 'ovn-sb.ovsschema'),
                 'remote_path': os.path.join(self.temp_dir, 'ovnsb_db.sock'),
                 'protocol': self.protocol,
                 'remote_ip': '127.0.0.1',
                 'remote_port': '0',
                 'pidfile': 'ovn-sb.pid',
                 'unixctl_path': os.path.join(self.temp_dir, 'ovnsb_db.ctl'),
                 'log_file_path': os.path.join(self.temp_dir, 'ovn_sb.log'),
                 'db_type': 'sb',
                 'connection': 'db:OVN_Southbound,SB_Global,connections',
                 'ctl_cmd': 'ovn-sbctl'})
        self.addCleanup(self.stop)
        self.start()

    def _init_ovsdb_pki(self):
        os.chdir(self.temp_dir)
        pki_init_cmd = [shutil.which('ovs-pki'), 'init',
                        '-d', self.temp_dir, '-l',
                        os.path.join(self.temp_dir, 'pki.log'), '--force']
        utils.execute(pki_init_cmd)
        pki_req_sign = [shutil.which('ovs-pki'), 'req+sign', 'ovn',
                        'controller', '-d', self.temp_dir, '-l',
                        os.path.join(self.temp_dir, 'pki.log'), '--force']
        utils.execute(pki_req_sign)

    def delete_dbs(self):
        for ovsdb in self.ovsdb_server_processes:
            try:
                os.remove(ovsdb['db_path'])
            except OSError:
                pass

    def start(self):
        pki_done = False
        for ovsdb_process in self.ovsdb_server_processes:
            # create the db from the schema using ovsdb-tool
            ovsdb_tool_cmd = [shutil.which('ovsdb-tool'),
                              'create', ovsdb_process['db_path'],
                              ovsdb_process['schema_path']]
            utils.execute(ovsdb_tool_cmd)

            # start the ovsdb-server
            ovsdb_server_cmd = [
                shutil.which('ovsdb-server'), '-vconsole:off',
                '--detach',
                '--pidfile=%s' % os.path.join(
                    self.temp_dir, ovsdb_process['pidfile']),
                '--log-file=%s' % (ovsdb_process['log_file_path']),
                '--remote=punix:%s' % (ovsdb_process['remote_path']),
                '--remote=%s' % (ovsdb_process['connection']),
                '--unixctl=%s' % (ovsdb_process['unixctl_path']),
                '--detach']
            if ovsdb_process['protocol'] == 'ssl':
                if not pki_done:
                    pki_done = True
                    self._init_ovsdb_pki()
                ovsdb_server_cmd.append('--private-key=%s' % self.private_key)
                ovsdb_server_cmd.append('--certificate=%s' % self.certificate)
                ovsdb_server_cmd.append('--ca-cert=%s' % self.ca_cert)
            ovsdb_server_cmd.append(ovsdb_process['db_path'])
            if self.debug:
                ovsdb_server_cmd.append('--verbose')
            obj, _ = utils.create_process(ovsdb_server_cmd)
            obj.communicate()

            conn_cmd = [shutil.which(ovsdb_process['ctl_cmd']),
                        '--db=unix:%s' % ovsdb_process['remote_path'],
                        'set-connection',
                        'p%s:%s:%s' % (ovsdb_process['protocol'],
                                       ovsdb_process['remote_port'],
                                       ovsdb_process['remote_ip']),
                        '--', 'set', 'connection', '.',
                        'inactivity_probe=60000']

            @tenacity.retry(wait=tenacity.wait_exponential(multiplier=0.1),
                            stop=tenacity.stop_after_delay(3), reraise=True)
            def _set_connection():
                utils.execute(conn_cmd)

            @tenacity.retry(
                wait=tenacity.wait_exponential(multiplier=0.1),
                stop=tenacity.stop_after_delay(10),
                reraise=True)
            def get_ovsdb_remote_port_retry(pid):
                process = psutil.Process(pid)
                for connect in process.connections():
                    if connect.status == 'LISTEN':
                        return connect.laddr[1]
                raise Exception(_("Could not find LISTEN port."))

            if ovsdb_process['protocol'] != 'unix':
                _set_connection()
                pid = self._get_pid_from_pidfile(ovsdb_process['pidfile'])
                ovsdb_process['remote_port'] = \
                    get_ovsdb_remote_port_retry(pid)

    def stop(self):
        for ovsdb_process in self.ovsdb_server_processes:
            try:
                stop_cmd = ['ovs-appctl', '-t', ovsdb_process['unixctl_path'],
                            'exit']
                utils.execute(stop_cmd)
            except Exception:
                pass

    def get_ovsdb_connection_path(self, db_type='nb'):
        for ovsdb_process in self.ovsdb_server_processes:
            if ovsdb_process['db_type'] == db_type:
                if ovsdb_process['protocol'] == 'unix':
                    return 'unix:' + ovsdb_process['remote_path']
                else:
                    return '%s:%s:%s' % (ovsdb_process['protocol'],
                                         ovsdb_process['remote_ip'],
                                         ovsdb_process['remote_port'])



# Copyright (C) 2014,2015 VA Linux Systems Japan K.K.
# Copyright (C) 2014 Fumihiko Kakuma <kakuma at valinux co jp>
# Copyright (C) 2014,2015 YAMAMOTO Takashi <yamamoto at valinux co jp>
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
from unittest import mock

from oslo_utils import importutils

from neutron.tests import base
from neutron.tests.unit.plugins.ml2.drivers.openvswitch.agent \
    import fake_oflib


_AGENT_PACKAGE = 'neutron.plugins.ml2.drivers.openvswitch.agent'
_AGENT_NAME = _AGENT_PACKAGE + '.ovs_neutron_agent'
_DVR_AGENT_NAME = ('neutron.plugins.ml2.drivers.openvswitch.agent.'
                   'ovs_dvr_neutron_agent')


class OVSAgentConfigTestBase(base.BaseTestCase):
    def setUp(self):
        super(OVSAgentConfigTestBase, self).setUp()
        self.mod_agent = importutils.import_module(_AGENT_NAME)
        self.mod_dvr_agent = importutils.import_module(_DVR_AGENT_NAME)


class OVSOSKenTestBase(OVSAgentConfigTestBase):

    _DRIVER_PACKAGE = _AGENT_PACKAGE + '.openflow.native'
    _BR_INT_CLASS = _DRIVER_PACKAGE + '.br_int.OVSIntegrationBridge'
    _BR_TUN_CLASS = _DRIVER_PACKAGE + '.br_tun.OVSTunnelBridge'
    _BR_PHYS_CLASS = _DRIVER_PACKAGE + '.br_phys.OVSPhysicalBridge'

    def setUp(self):
        self.fake_oflib_of = fake_oflib.patch_fake_oflib_of()
        self.fake_oflib_of.start()
        self.addCleanup(self.fake_oflib_of.stop)
        super(OVSOSKenTestBase, self).setUp()
        conn_patcher = mock.patch(
            'neutron.agent.ovsdb.impl_idl._connection')
        conn_patcher.start()
        self.addCleanup(conn_patcher.stop)
        self.br_int_cls = importutils.import_class(self._BR_INT_CLASS)
        self.br_phys_cls = importutils.import_class(self._BR_PHYS_CLASS)
        self.br_tun_cls = importutils.import_class(self._BR_TUN_CLASS)
        os_ken_app = mock.Mock()
        self.br_int_cls = functools.partial(self.br_int_cls,
                                            os_ken_app=os_ken_app)
        self.br_phys_cls = functools.partial(self.br_phys_cls,
                                             os_ken_app=os_ken_app)
        self.br_tun_cls = functools.partial(self.br_tun_cls,
                                            os_ken_app=os_ken_app)

    def _bridge_classes(self):
        return {
            'br_int': self.br_int_cls,
            'br_phys': self.br_phys_cls,
            'br_tun': self.br_tun_cls,
        }



#!/usr/bin/env python3
# Copyright 2017 OVH SAS
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys
from unittest import mock

from oslo_config import cfg

from neutron.agent.common import ovs_lib
from neutron.agent.common import polling
from neutron.agent.l2.extensions import qos as qos_extension
from neutron.common import config
from neutron.tests.common.agents import ovs_agent


def monkeypatch_qos():
    mock.patch.object(ovs_lib.OVSBridge, 'clear_bandwidth_qos').start()
    if "qos" in cfg.CONF.service_plugins:
        mock.patch.object(qos_extension.QosAgentExtension,
                          '_process_reset_port').start()


def monkeypatch_event_filtering():
    def filter_bridge_names(br_names):
        if 'trunk' in cfg.CONF.service_plugins:
            return []
        return br_names

    polling.filter_bridge_names = filter_bridge_names


def main():
    # TODO(slaweq): this monkepatch will not be necessary when
    # https://review.opendev.org/#/c/506722/ will be merged and ovsdb-server
    # ovs-vswitchd processes for each test will be isolated in separate
    # namespace
    config.register_common_config_options()
    monkeypatch_qos()
    monkeypatch_event_filtering()
    ovs_agent.main()


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2014 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from neutron._i18n import _
from neutron.conf.plugins.ml2.drivers.mech_sriov import agent_common as \
     agent_common_config
from neutron.plugins.ml2.drivers.agent import config as cagt_config  # noqa


def parse_exclude_devices(exclude_list):
    """Parse Exclude devices list

    parses excluded device list in the form:
    dev_name:pci_dev_1;pci_dev_2
    @param exclude list: list of string pairs in "key:value" format
                        the key part represents the network device name
                        the value part is a list of PCI slots separated by ";"
    """
    exclude_mapping = {}
    for dev_mapping in exclude_list:
        try:
            dev_name, exclude_devices = dev_mapping.split(":", 1)
        except ValueError:
            raise ValueError(_("Invalid mapping: '%s'") % dev_mapping)
        dev_name = dev_name.strip()
        if not dev_name:
            raise ValueError(_("Missing key in mapping: '%s'") % dev_mapping)
        if dev_name in exclude_mapping:
            raise ValueError(_("Device %(dev_name)s in mapping: %(mapping)s "
                               "not unique") % {'dev_name': dev_name,
                                                'mapping': dev_mapping})
        exclude_devices_list = exclude_devices.split(";")
        exclude_devices_set = set()
        for dev in exclude_devices_list:
            dev = dev.strip()
            if dev:
                exclude_devices_set.add(dev)
        exclude_mapping[dev_name] = exclude_devices_set
    return exclude_mapping


agent_common_config.register_agent_sriov_nic_opts()



# Copyright 2012 VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from unittest import mock

from neutron.plugins.ml2.drivers.openvswitch.agent \
    import ovs_agent_extension_api as ovs_ext_agt

from neutron.tests.unit.plugins.ml2.drivers.openvswitch.agent \
    .openflow.native import ovs_bridge_test_base as native_ovs_bridge_test_base

from neutron.tests.unit.plugins.ml2.drivers.openvswitch.agent \
    import ovs_test_base


class TestOVSAgentExtensionAPI(ovs_test_base.OVSOSKenTestBase):

    def setUp(self):
        super(TestOVSAgentExtensionAPI, self).setUp()
        self.br_int = self.br_int_cls("br-int")
        self.br_tun = self.br_tun_cls("br-tun")
        self.br_phys = {'br-phys1': self.br_phys_cls('br-phys1'),
                        'br-phys2': self.br_phys_cls('br-phys2')}

    def _test_bridge(self, orig_bridge, new_bridge):
        self.assertIsNotNone(new_bridge)
        self.assertEqual(orig_bridge.br_name, new_bridge.br_name)
        self.assertIn(new_bridge._default_cookie,
                      orig_bridge.reserved_cookies)
        self.assertNotEqual(orig_bridge._default_cookie,
                            new_bridge._default_cookie)

    def test_request_int_br(self):
        agent_extension_api = ovs_ext_agt.OVSAgentExtensionAPI(
            self.br_int, self.br_tun, {'phys': self.br_phys['br-phys1']})
        new_int_br = agent_extension_api.request_int_br()
        self._test_bridge(self.br_int, new_int_br)

    def test_request_tun_br(self):
        agent_extension_api = ovs_ext_agt.OVSAgentExtensionAPI(
            self.br_int, self.br_tun, {'phys': self.br_phys['br-phys1']})
        new_tun_br = agent_extension_api.request_tun_br()
        self._test_bridge(self.br_tun, new_tun_br)

    def test_request_tun_br_tunneling_disabled(self):
        agent_extension_api = ovs_ext_agt.OVSAgentExtensionAPI(
            self.br_int, None, {'phys': self.br_phys['br-phys1']})
        self.assertIsNone(agent_extension_api.request_tun_br())

    def test_request_phys_brs(self):
        agent_extension_api = ovs_ext_agt.OVSAgentExtensionAPI(
            self.br_int, self.br_tun,
            {'phys1': self.br_phys['br-phys1'],
             'phys2': self.br_phys['br-phys2']})
        for phys_br in agent_extension_api.request_phy_brs():
            self._test_bridge(self.br_phys[phys_br.br_name], phys_br)

    def test_request_physical_br(self):
        agent_extension_api = ovs_ext_agt.OVSAgentExtensionAPI(
            self.br_int, self.br_tun,
            {'phys1': self.br_phys['br-phys1'],
             'phys2': self.br_phys['br-phys2']})
        phys_br = agent_extension_api.request_physical_br('phys1')
        self._test_bridge(self.br_phys[phys_br.br_name], phys_br)


class TestOVSCookieBridgeOSKen(native_ovs_bridge_test_base.OVSBridgeTestBase):

    def setUp(self):
        super(TestOVSCookieBridgeOSKen, self).setUp()
        self.setup_bridge_mock('br-int', self.br_int_cls)
        self.tested_bridge = ovs_ext_agt.OVSCookieBridge(self.br)

    def test_cookie(self):
        self.assertNotEqual(self.br._default_cookie,
                            self.tested_bridge._default_cookie)

    def test_reserved(self):
        self.assertIn(self.tested_bridge._default_cookie,
                      self.br.reserved_cookies)

    def test_install_drop(self):
        priority = 99
        in_port = 666
        self.tested_bridge.install_drop(priority=priority,
                                        in_port=in_port)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            mock.call._send_msg(
                ofpp.OFPFlowMod(
                    dp,
                    # this is the interesting part of the check:
                    cookie=self.tested_bridge._default_cookie,
                    instructions=[],
                    match=ofpp.OFPMatch(in_port=in_port),
                    priority=priority,
                    table_id=0),
                active_bundle=None),
        ]
        self.assertEqual(expected, self.mock.mock_calls)



# Copyright (c) 2014 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
The module provides all database models at current HEAD.

Its purpose is to create comparable metadata with current database schema.
Based on this comparison database can be healed with healing migration.

"""

import os.path

from neutron_lib.db import model_base

from neutron.common import utils
from neutron.db import agentschedulers_db  # noqa
from neutron.db.extra_dhcp_opt import models as edo_models  # noqa
from neutron.db import l3_dvrscheduler_db  # noqa
from neutron.db import l3_gwmode_db  # noqa
from neutron.db import models
from neutron.db import models_v2  # noqa
from neutron.db.port_security import models as ps_models  # noqa
from neutron.db.qos import models as qos_models  # noqa
from neutron.db.quota import models as quota_models  # noqa
from neutron.db import rbac_db_models  # noqa
from neutron.ipam.drivers.neutrondb_ipam import db_models  # noqa
from neutron.plugins.ml2 import models as ml2_models  # noqa
from neutron.services.auto_allocate import models as aa_models  # noqa
from neutron.services.trunk import models as trunk_models  # noqa


utils.import_modules_recursively(os.path.dirname(models.__file__))


def get_metadata():
    return model_base.BASEV2.metadata



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
from neutron_lib.api.definitions import uplink_status_propagation as apidef
from neutron_lib.db import api as db_api
from neutron_lib.db import resource_extend

from neutron.db import db_base_plugin_v2
from neutron.db import uplink_status_propagation_db as usp_db
from neutron.tests.unit.db import test_db_base_plugin_v2


class UplinkStatusPropagationExtensionTestPlugin(
        db_base_plugin_v2.NeutronDbPluginV2,
        usp_db.UplinkStatusPropagationMixin):
    """Test plugin to mixin the uplink status propagation extension.
    """

    supported_extension_aliases = [apidef.ALIAS]

    @staticmethod
    @resource_extend.extends([apidef.COLLECTION_NAME])
    def _extend_network_project_default(port_res, port_db):
        return usp_db.UplinkStatusPropagationMixin._extend_port_dict(
            port_res, port_db)

    def create_port(self, context, port):
        with db_api.CONTEXT_WRITER.using(context):
            new_port = super(UplinkStatusPropagationExtensionTestPlugin,
                            self).create_port(context, port)
            # Update the propagate_uplink_status in the database
            p = port['port']
            if 'propagate_uplink_status' not in p:
                p['propagate_uplink_status'] = False
            self._process_create_port(context, p, new_port)
        return new_port


@ddt.ddt
class UplinkStatusPropagationExtensionTestCase(
         test_db_base_plugin_v2.NeutronDbPluginV2TestCase):
    """Test API extension propagate_uplink_status attributes.
    """

    def setUp(self):
        plugin = ('neutron.tests.unit.extensions.test_uplink_status_'
                  'propagation.UplinkStatusPropagationExtensionTestPlugin')
        super(UplinkStatusPropagationExtensionTestCase,
              self).setUp(plugin=plugin)

    @ddt.data(True, False)
    def test_create_port_propagate_uplink_status(
            self, propagate_uplink_status):
        name = 'propagate_uplink_status'
        keys = [('name', name), ('admin_state_up', True),
                ('status', self.port_create_status),
                ('propagate_uplink_status', propagate_uplink_status)]
        with self.port(name=name,
                       propagate_uplink_status=propagate_uplink_status
                       ) as port:
            for k, v in keys:
                self.assertEqual(v, port['port'][k])



# Copyright 2015 OpenStack Foundation
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_config import cfg

from neutron._i18n import _
from neutron.conf.agent import common
from neutron.conf.agent.metadata import config as meta_conf

DHCP_AGENT_OPTS = [
    cfg.IntOpt('resync_interval', default=5,
               help=_("The DHCP agent will resync its state with Neutron to "
                      "recover from any transient notification or RPC errors. "
                      "The interval is the maximum number of seconds between "
                      "attempts. The resync can be done more often based on "
                      "the events triggered.")),
    cfg.IntOpt('resync_throttle', default=1,
               help=_("Throttle the number of resync state events between the "
                      "local DHCP state and Neutron to only once per "
                      "'resync_throttle' seconds. The value of throttle "
                      "introduces a minimum interval between resync state "
                      "events. Otherwise the resync may end up in a "
                      "busy-loop. The value must be less than "
                      "resync_interval.")),
    cfg.StrOpt('dhcp_driver',
               default='neutron.agent.linux.dhcp.Dnsmasq',
               help=_("The driver used to manage the DHCP server.")),
    cfg.BoolOpt('enable_isolated_metadata', default=False,
                help=_("The DHCP server can assist with providing metadata "
                       "support on isolated networks. Setting this value to "
                       "True will cause the DHCP server to append specific "
                       "host routes to the DHCP request. The metadata service "
                       "will only be activated when the subnet does not "
                       "contain any router port. The guest instance must be "
                       "configured to request host routes via DHCP (Option "
                       "121). This option does not have any effect when "
                       "force_metadata is set to True.")),
    cfg.BoolOpt('force_metadata', default=False,
                help=_("In some cases the Neutron router is not present to "
                       "provide the metadata IP but the DHCP server can be "
                       "used to provide this info. Setting this value will "
                       "force the DHCP server to append specific host routes "
                       "to the DHCP request. If this option is set, then the "
                       "metadata service will be activated for all of the "
                       "networks.")),
    cfg.BoolOpt('enable_metadata_network', default=False,
                help=_("Allows for serving metadata requests coming from a "
                       "dedicated metadata access network whose CIDR is "
                       "169.254.169.254/16 (or larger prefix), and is "
                       "connected to a Neutron router from which the VMs send "
                       "metadata:1 request. In this case DHCP Option 121 will "
                       "not be injected in VMs, as they will be able to reach "
                       "169.254.169.254 through a router. This option "
                       "requires enable_isolated_metadata = True.")),
    cfg.IntOpt('num_sync_threads', default=4,
               help=_('Number of threads to use during sync process. '
                      'Should not exceed connection pool size configured on '
                      'server.')),
    cfg.IntOpt('bulk_reload_interval', default=0, min=0,
               help=_('Time to sleep between reloading the DHCP allocations. '
                      'This will only be invoked if the value is not 0. '
                      'If a network has N updates in X seconds then '
                      'it will reload once and not N times.')),
]

DHCP_OPTS = [
    cfg.StrOpt('dhcp_confs',
               default='$state_path/dhcp',
               help=_('Location to store DHCP server config files.')),
]

DNSMASQ_OPTS = [
    cfg.StrOpt('dnsmasq_config_file',
               default='',
               help=_('Override the default dnsmasq settings '
                      'with this file.')),
    cfg.ListOpt('dnsmasq_dns_servers',
                default=[],
                help=_('Comma-separated list of the DNS servers which will be '
                       'used as forwarders.')),
    cfg.StrOpt('dnsmasq_base_log_dir',
               help=_("Base log dir for dnsmasq logging. "
                      "The log contains DHCP and DNS log information and "
                      "is useful for debugging issues with either DHCP or "
                      "DNS. If this section is null, disable dnsmasq log.")),
    cfg.BoolOpt('dnsmasq_local_resolv', default=False,
                help=_("Enables the dnsmasq service to provide name "
                       "resolution for instances via DNS resolvers on the "
                       "host running the DHCP agent. Effectively removes the "
                       "'--no-resolv' option from the dnsmasq process "
                       "arguments. Adding custom DNS resolvers to the "
                       "'dnsmasq_dns_servers' option disables this feature.")),
    cfg.IntOpt(
        'dnsmasq_lease_max',
        default=(2 ** 24),
        help=_('Limit number of leases to prevent a denial-of-service.')),
    cfg.BoolOpt('dhcp_broadcast_reply', default=False,
                help=_("Use broadcast in DHCP replies.")),
    cfg.BoolOpt('dnsmasq_enable_addr6_list', default=False,
                help=_("Enable dhcp-host entry with list of addresses when "
                       "port has multiple IPv6 addresses in the same subnet."))
]


def register_agent_dhcp_opts(cfg=cfg.CONF):
    cfg.register_opts(DHCP_AGENT_OPTS)
    cfg.register_opts(DHCP_OPTS)
    cfg.register_opts(DNSMASQ_OPTS)
    cfg.register_opts(common.DHCP_PROTOCOL_OPTS)
    meta_conf.register_meta_conf_opts(meta_conf.METADATA_RATE_LIMITING_OPTS,
                                      cfg=cfg,
                                      group=meta_conf.RATE_LIMITING_GROUP)



# Copyright 2020 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from os import path
import re

from eventlet.green import subprocess
from neutron_lib.utils import helpers
from oslo_concurrency import processutils
from oslo_utils import fileutils

from neutron import privileged


NETSTAT_PIDS_REGEX = re.compile(r'.* (?P<pid>\d{2,6})/.*')


@privileged.default.entrypoint
def find_listen_pids_namespace(namespace):
    return _find_listen_pids_namespace(namespace)


def _find_listen_pids_namespace(namespace):
    """Retrieve a list of pids of listening processes within the given netns

    This method is implemented separately to allow unit testing.
    """
    pids = set()
    cmd = ['ip', 'netns', 'exec', namespace, 'netstat', '-nlp']
    output = processutils.execute(*cmd)
    for line in output[0].splitlines():
        m = NETSTAT_PIDS_REGEX.match(line)
        if m:
            pids.add(m.group('pid'))
    return list(pids)


@privileged.default.entrypoint
def delete_if_exists(_path, remove=os.unlink):
    fileutils.delete_if_exists(_path, remove=remove)


@privileged.default.entrypoint
def execute_process(cmd, _process_input, addl_env):
    obj, cmd = _create_process(cmd, addl_env=addl_env)
    _stdout, _stderr = obj.communicate(_process_input)
    returncode = obj.returncode
    obj.stdin.close()
    _stdout = helpers.safe_decode_utf8(_stdout)
    _stderr = helpers.safe_decode_utf8(_stderr)
    return _stdout, _stderr, returncode


def _addl_env_args(addl_env):
    """Build arguments for adding additional environment vars with env"""

    # NOTE (twilson) If using rootwrap, an EnvFilter should be set up for the
    # command instead of a CommandFilter.
    if addl_env is None:
        return []
    return ['env'] + ['%s=%s' % pair for pair in addl_env.items()]


def _create_process(cmd, addl_env=None):
    """Create a process object for the given command.

    The return value will be a tuple of the process object and the
    list of command arguments used to create it.
    """
    cmd = list(map(str, _addl_env_args(addl_env) + list(cmd)))
    obj = subprocess.Popen(cmd, shell=False, stdin=subprocess.PIPE,
                           stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return obj, cmd


@privileged.default.entrypoint
def path_exists(_path):
    return path.exists(_path)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib import constants as n_const
from neutron_lib import context
from neutron_lib import exceptions
from neutron_lib.services.logapi import constants as log_const

from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils
from neutron.services.logapi.common import exceptions as log_exc
from neutron.tests.functional import base as functional_base


class LogApiTestCaseBase(functional_base.TestOVNFunctionalBase):
    def setUp(self):
        super().setUp()
        self.log_driver = self.mech_driver.log_driver
        self._check_is_supported()
        self.ctxt = context.Context('admin', self._tenant_id)

    def _check_is_supported(self):
        if not self.log_driver.network_logging_supported(self.nb_api):
            self.skipTest("The current OVN version does not offer support "
                          "for neutron network log functionality.")
        self.assertIsNotNone(self.log_plugin)

    def _log_data(self, sg_id=None, port_id=None, enabled=True):
        log_data = {'project_id': self.ctxt.project_id,
                    'resource_type': 'security_group',
                    'description': 'test net log',
                    'name': 'logme',
                    'enabled': enabled,
                    'event': log_const.ALL_EVENT}
        if sg_id:
            log_data['resource_id'] = sg_id
        if port_id:
            log_data['target_id'] = port_id
        return {'log': log_data}


class LogApiTestCaseSimple(LogApiTestCaseBase):
    def test_basic_get(self):
        log_obj = self.log_plugin.create_log(self.ctxt, self._log_data())
        self.assertIsNotNone(log_obj)
        log_obj_get = self.log_plugin.get_log(self.ctxt, log_obj['id'])
        self.assertEqual(log_obj, log_obj_get)
        log_obj2 = self.log_plugin.create_log(self.ctxt, self._log_data())
        self.assertIsNotNone(log_obj2)
        log_objs_get = self.log_plugin.get_logs(self.ctxt)
        log_objs_ids = {x['id'] for x in log_objs_get}
        self.assertEqual({log_obj['id'], log_obj2['id']}, log_objs_ids)

    def test_log_ovn_unsupported(self):
        with mock.patch.object(self.log_driver, 'network_logging_supported',
                        return_value=False) as supported_mock:
            log_data = {'log': {'resource_type': 'security_group',
                                'enabled': True}}
            self.assertRaises(exceptions.DriverCallError,
                              self.log_plugin.create_log,
                              self.ctxt, log_data)
            supported_mock.assert_called_once()


class LogApiTestCaseComplex(LogApiTestCaseBase):
    def setUp(self):
        super().setUp()
        self._prepare_env()

    def _prepare_env(self):
        self.net = self._create_network(
            self.fmt, 'private', admin_state_up=True).json['network']['id']
        self.subnet = self._create_subnet(
            self.fmt, self.net, '10.0.0.0/24', enable_dhcp=False).json[
            'subnet']['id']

        self.sg1 = self._create_security_group('test_sg1_ssh')
        self.sg2 = self._create_security_group('test_sg2_http')
        self.sg3 = self._create_security_group('test_sg3_telnet_ssh')
        self.sg1rs = [self._create_security_group_rule(self.sg1, 22)]
        self.sg2rs = [self._create_security_group_rule(self.sg2, 80)]
        self.sg3rs = [self._create_security_group_rule(self.sg3, 23),
                      self._create_security_group_rule(self.sg3, 22)]
        self.sgs = [self.sg1, self.sg2, self.sg3]
        self.sgrs = self.sg1rs + self.sg2rs + self.sg3rs

        self.port1_sgs = [self.sg1]
        self.port1_sgrs = self.sg1rs
        self.port1 = self._create_port(self.fmt, self.net,
                                       security_groups=self.port1_sgs)
        self.port2_sgs = [self.sg2, self.sg3]
        self.port2_sgrs = self.sg2rs + self.sg3rs
        self.port2 = self._create_port(self.fmt, self.net,
                                       security_groups=self.port2_sgs)
        self.port3_sgs = [self.sg1, self.sg3]
        self.port3_sgrs = self.sg1rs + self.sg3rs
        self.port3 = self._create_port(self.fmt, self.net,
                                       security_groups=self.port3_sgs)

    def _create_port(self, name, net_id, security_groups):
        data = {'port': {'name': name,
                         'network_id': net_id,
                         'security_groups': security_groups}}
        req = self.new_create_request('ports', data, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['port']['id']

    def _create_security_group(self, name):
        data = {'security_group': {'name': name}}
        req = self.new_create_request('security-groups', data, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['security_group']['id']

    def _create_security_group_rule(self, sg_id, tcp_port):
        data = {'security_group_rule': {'security_group_id': sg_id,
                                        'direction': 'ingress',
                                        'protocol': n_const.PROTO_NAME_TCP,
                                        'ethertype': n_const.IPv4,
                                        'port_range_min': tcp_port,
                                        'port_range_max': tcp_port}}
        req = self.new_create_request('security-group-rules', data, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['security_group_rule']['id']

    def _find_security_group_row_by_id(self, sg_id):
        for row in self.nb_api._tables['Port_Group'].rows.values():
            if row.name == utils.ovn_port_group_name(sg_id):
                return row

    def _find_security_group_rule_row_by_id(self, sgr_id):
        for row in self.nb_api._tables['ACL'].rows.values():
            if (row.external_ids.get(
                    ovn_const.OVN_SG_RULE_EXT_ID_KEY) == sgr_id):
                return row

    def _check_acl_log(self, sgr, is_enabled=True):
        acl = self._find_security_group_rule_row_by_id(sgr)
        self.assertIsNotNone(acl)
        self.assertEqual(is_enabled, acl.log)
        if hasattr(acl, "label"):
            # Here we compare if there is a name because the log can be
            # disabled but disabling a log would not take out the properties
            # attached to it.
            if acl.name:
                self.assertNotEqual(0, acl.label)
                self.assertEqual("true", acl.options.get("log-related"))
            else:
                self.assertEqual(0, acl.label)
                self.assertIsNone(acl.options.get("log-related"))
        return acl

    def _check_acl_log_drop(self, is_enabled=True):
        acls = self.nb_api.get_port_group(
            ovn_const.OVN_DROP_PORT_GROUP_NAME).acls
        self.assertTrue(acls)
        for acl in acls:
            self.assertEqual(is_enabled, acl.log)
        return acls

    def _check_sgrs(self, sgrs=None, is_enabled=True):
        if not sgrs:
            sgrs = self.sgrs
        for sgr in sgrs:
            self._check_acl_log(sgr, is_enabled)

    def test_add_and_remove(self):
        self._check_sgrs(is_enabled=False)
        self.assertEqual([],
                         self.nb_api.meter_list().execute(check_error=True))

        log_obj = self.log_plugin.create_log(self.ctxt, self._log_data())
        for sgr in self.sgrs:
            acl = self._check_acl_log(sgr)
            self.assertEqual(utils.ovn_name(log_obj['id']), acl.name[0])
            meter = self.nb_api.meter_get(acl.meter[0]).execute(
                check_error=True)
            self.assertEqual([True], meter.fair)
            self.assertEqual('pktps', meter.unit)
            self.assertEqual(1, len(meter.bands))
            self.assertEqual({ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY:
                              log_const.LOGGING_PLUGIN}, meter.external_ids)

        self.log_plugin.delete_log(self.ctxt, log_obj['id'])
        self._check_sgrs(is_enabled=False)
        self.assertEqual([],
                         self.nb_api.meter_list().execute(check_error=True))

        log_objs = []
        for sg in self.sgs:
            log_data = self._log_data(sg_id=sg)
            log_objs.append(self.log_plugin.create_log(self.ctxt, log_data))
        self.assertEqual(len(log_objs),
                         len(self.log_plugin.get_logs(self.ctxt)))
        self._check_sgrs(is_enabled=True)

        # Attempt to delete non-existing row
        self.assertRaises(log_exc.LogResourceNotFound,
                          self.log_plugin.delete_log,
                          self.ctxt, log_obj['id'])

        self.log_plugin.delete_log(self.ctxt, log_objs[1]['id'])
        self._check_sgrs(sgrs=self.sg1rs, is_enabled=True)
        self._check_sgrs(sgrs=self.sg2rs, is_enabled=False)
        self._check_sgrs(sgrs=self.sg3rs, is_enabled=True)

        self.log_plugin.delete_log(self.ctxt, log_objs[2]['id'])
        self._check_sgrs(sgrs=self.sg1rs, is_enabled=True)
        self._check_sgrs(sgrs=self.sg2rs, is_enabled=False)
        self._check_sgrs(sgrs=self.sg3rs, is_enabled=False)

        self.log_plugin.delete_log(self.ctxt, log_objs[0]['id'])
        self.assertEqual([], self.log_plugin.get_logs(self.ctxt))
        self._check_sgrs(is_enabled=False)

        # Attempt to delete from empty table
        self.assertRaises(log_exc.LogResourceNotFound,
                          self.log_plugin.delete_log,
                          self.ctxt, log_objs[0]['id'])

    def test_update_all(self):
        # Note: only these fields are supported for update:
        # openstack network log set [-h] [--description <description>]
        # [--enable | --disable] [--name <name>] <network-log>

        log_data = self._log_data()
        log_obj = self.log_plugin.create_log(self.ctxt, log_data)
        self._check_sgrs()

        log_data['log']['name'] = 'logme-nay'
        log_data['log']['enabled'] = False
        self.log_plugin.update_log(self.ctxt, log_obj['id'], log_data)
        self._check_sgrs(is_enabled=False)

        log_data['log']['name'] = 'logme-yay'
        log_data['log']['description'] = 'logs are a beautiful thing'
        log_data['log']['enabled'] = True
        self.log_plugin.update_log(self.ctxt, log_obj['id'], log_data)
        self._check_sgrs()

    def test_update_one_sg(self):
        log_data = self._log_data(sg_id=self.sg2, enabled=False)
        log_obj = self.log_plugin.create_log(self.ctxt, log_data)
        self._check_sgrs(is_enabled=False)

        log_data['log']['enabled'] = True
        self.log_plugin.update_log(self.ctxt, log_obj['id'], log_data)
        self._check_sgrs(sgrs=self.sg1rs, is_enabled=False)
        self._check_sgrs(sgrs=self.sg2rs, is_enabled=True)
        self._check_sgrs(sgrs=self.sg3rs, is_enabled=False)

    def test_overlap_net_logs(self):
        log_data1 = self._log_data(sg_id=self.sg3, port_id=self.port3)
        log_obj1 = self.log_plugin.create_log(self.ctxt, log_data1)
        self._check_sgrs(sgrs=self.sg1rs, is_enabled=False)
        self._check_sgrs(sgrs=self.sg2rs, is_enabled=False)
        self._check_sgrs(sgrs=self.sg3rs, is_enabled=True)

        log_data2 = self._log_data(port_id=self.port2)
        log_obj2 = self.log_plugin.create_log(self.ctxt, log_data2)
        self._check_sgrs(sgrs=self.sg1rs, is_enabled=False)

        # port 2 uses sg2 and sg3. However, sg3 is in use by log_obj1
        # so only acls for 2 would be associated with log_obj2
        for sgr in self.sg2rs:
            acl = self._check_acl_log(sgr)
            self.assertEqual(utils.ovn_name(log_obj2['id']), acl.name[0])
        for sgr in self.sg3rs:
            acl = self._check_acl_log(sgr)
            self.assertEqual(utils.ovn_name(log_obj1['id']), acl.name[0])

        # Next, delete log_obj1 and make sure that lob_obj2 gets to
        # claim what it could not use before
        self.log_plugin.delete_log(self.ctxt, log_obj1['id'])
        self._check_sgrs(sgrs=self.sg1rs, is_enabled=False)
        for sgr in self.sg2rs + self.sg3rs:
            acl = self._check_acl_log(sgr)
            self.assertEqual(utils.ovn_name(log_obj2['id']), acl.name[0])

        # Delete log_obj2 and ensure that logs are off and meter is no
        # longer used
        self.log_plugin.delete_log(self.ctxt, log_obj2['id'])
        self._check_sgrs(is_enabled=False)
        self.assertEqual([],
                         self.nb_api.meter_list().execute(check_error=True))

    def _add_logs_then_remove(self, event1, event2, sg=None, sgrs=None):
        # Events were previously not correctly applied on ACLs. This test
        # ensures that each event log only the necessary acls
        drop_true_events = (log_const.DROP_EVENT, log_const.ALL_EVENT)
        accept_true_events = (log_const.ALL_EVENT, log_const.ACCEPT_EVENT)
        # Check there are no acls with their logging active
        self._check_sgrs(sgrs=sgrs, is_enabled=False)
        self._check_acl_log_drop(is_enabled=False)

        # Add first log object
        log_data1 = self._log_data(sg_id=sg)
        log_data1['log']['event'] = event1
        log_obj1 = self.log_plugin.create_log(self.ctxt, log_data1)
        self._check_acl_log_drop(is_enabled=event1 in drop_true_events)
        self._check_sgrs(sgrs=sgrs, is_enabled=event1 in accept_true_events)

        # Add second log object
        log_data2 = self._log_data(sg_id=sg)
        log_data2['log']['event'] = event2
        log_obj2 = self.log_plugin.create_log(self.ctxt, log_data2)
        self._check_acl_log_drop(is_enabled=(event1 in drop_true_events or
            event2 in drop_true_events))
        self._check_sgrs(sgrs=sgrs, is_enabled=(event1 in accept_true_events or
            event2 in accept_true_events))

        # Delete second log object
        self.log_plugin.delete_log(self.ctxt, log_obj2['id'])
        self._check_acl_log_drop(is_enabled=event1 in drop_true_events)
        self._check_sgrs(sgrs=sgrs, is_enabled=event1 in accept_true_events)

        # Delete first log object
        self.log_plugin.delete_log(self.ctxt, log_obj1['id'])
        self._check_sgrs(sgrs=sgrs, is_enabled=False)
        self._check_acl_log_drop(is_enabled=False)

    def test_events_all_sg(self):
        self._add_logs_then_remove(log_const.DROP_EVENT, log_const.ALL_EVENT)
        self._add_logs_then_remove(
            log_const.ACCEPT_EVENT, log_const.DROP_EVENT)
        self._add_logs_then_remove(
            log_const.DROP_EVENT, log_const.ACCEPT_EVENT)

    def test_events_one_sg(self):
        self._add_logs_then_remove(log_const.DROP_EVENT, log_const.ALL_EVENT,
                                   sg=self.sg1, sgrs=self.sg1rs)
        self._add_logs_then_remove(
            log_const.ACCEPT_EVENT, log_const.DROP_EVENT, sg=self.sg2,
            sgrs=self.sg2rs)
        self._add_logs_then_remove(
            log_const.DROP_EVENT, log_const.ACCEPT_EVENT, sg=self.sg3,
            sgrs=self.sg3rs)

    def test_disable_logs(self):
        # This test ensures that acls are correctly disabled when having
        # multiple log objects.

        # Check there are no acls with their logging active
        sgrs = self.sg1rs
        self._check_sgrs(sgrs, is_enabled=False)
        self._check_acl_log_drop(is_enabled=False)

        # Add accept log object
        log_data1 = self._log_data(sg_id=self.sg1)
        event1 = log_const.ACCEPT_EVENT
        log_data1['log']['event'] = event1
        log_obj1 = self.log_plugin.create_log(self.ctxt, log_data1)
        self._check_acl_log_drop(is_enabled=False)
        self._check_sgrs(sgrs=sgrs, is_enabled=True)

        # Add drop log object
        log_data2 = self._log_data(sg_id=self.sg1)
        event2 = log_const.DROP_EVENT
        log_data2['log']['event'] = event2
        log_obj2 = self.log_plugin.create_log(self.ctxt, log_data2)
        self._check_acl_log_drop(is_enabled=True)
        self._check_sgrs(sgrs=sgrs, is_enabled=True)

        # Disable drop log object and check it worked correctly
        log_data2['log']['enabled'] = False
        self.log_plugin.update_log(self.ctxt, log_obj2['id'], log_data2)
        self._check_acl_log_drop(is_enabled=False)
        self._check_sgrs(sgrs=sgrs, is_enabled=True)

        # Enable drop log and create all log object
        log_data2['log']['enabled'] = True
        self.log_plugin.update_log(self.ctxt, log_obj2['id'], log_data2)
        self._check_acl_log_drop(is_enabled=True)
        self._check_sgrs(sgrs=sgrs, is_enabled=True)

        log_data3 = self._log_data(sg_id=self.sg1)
        log_data3['log']['event'] = log_const.ALL_EVENT
        log_obj3 = self.log_plugin.create_log(self.ctxt, log_data3)
        self._check_sgrs(sgrs=sgrs, is_enabled=True)
        self._check_acl_log_drop(is_enabled=True)

        # Disable all log object and check all acls are still enabled (because
        # of the other objects)
        log_data3['log']['enabled'] = False
        self.log_plugin.update_log(self.ctxt, log_obj3['id'], log_data3)
        self._check_sgrs(sgrs=sgrs, is_enabled=True)
        self._check_acl_log_drop(is_enabled=True)

        # Disable accept log object and only drop traffic gets logged
        log_data1['log']['enabled'] = False
        self.log_plugin.update_log(self.ctxt, log_obj1['id'], log_data1)
        self._check_sgrs(sgrs=sgrs, is_enabled=False)
        self._check_acl_log_drop(is_enabled=True)



# Copyright 2014 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from oslo_config import cfg

from neutron._i18n import _

DEFAULT_DEVICE_MAPPINGS = []
DEFAULT_EXCLUDE_DEVICES = []

sriov_nic_opts = [
    cfg.ListOpt('physical_device_mappings',
                default=DEFAULT_DEVICE_MAPPINGS,
                help=_("Comma-separated list of "
                       "<physical_network>:<network_device> tuples mapping "
                       "physical network names to the agent's node-specific "
                       "physical network device interfaces of SR-IOV physical "
                       "function to be used for VLAN networks. All physical "
                       "networks listed in network_vlan_ranges on the server "
                       "should have mappings to appropriate interfaces on "
                       "each agent.")),
    cfg.ListOpt('exclude_devices',
                default=DEFAULT_EXCLUDE_DEVICES,
                help=_("Comma-separated list of "
                       "<network_device>:<vfs_to_exclude> tuples, mapping "
                       "network_device to the agent's node-specific list of "
                       "virtual functions that should not be used for virtual "
                       "networking. vfs_to_exclude is a semicolon-separated "
                       "list of virtual functions to exclude from "
                       "network_device. The network_device in the mapping "
                       "should appear in the physical_device_mappings "
                       "list.")),
    cfg.ListOpt('resource_provider_bandwidths',
                default=[],
                help=_("Comma-separated list of "
                       "<network_device>:<egress_bw>:<ingress_bw> tuples, "
                       "showing the available bandwidth for the given device "
                       "in the given direction. The direction is meant from "
                       "VM perspective. Bandwidth is measured in kilobits per "
                       "second (kbps). The device must appear in "
                       "physical_device_mappings as the value. But not all "
                       "devices in physical_device_mappings must be listed "
                       "here. For a device not listed here we neither create "
                       "a resource provider in placement nor report "
                       "inventories against. An omitted direction means we do "
                       "not report an inventory for the corresponding "
                       "class.")),
    cfg.DictOpt('resource_provider_hypervisors',
                default={},
                help=_("Mapping of network devices to hypervisors: "
                       "<network_device>:<hypervisor>,... "
                       "hypervisor name is used to locate the parent of the "
                       "resource provider tree. Only needs to be set in the "
                       "rare case when the hypervisor name is different from "
                       "the resource_provider_default_hypervisor config "
                       "option value as known by the nova-compute managing "
                       "that hypervisor.")),
    cfg.StrOpt('resource_provider_default_hypervisor',
               help=_("The default hypervisor name used to locate the parent "
                      "of the resource provider. If this option is not set, "
                      "canonical name is used")),
    cfg.DictOpt('resource_provider_inventory_defaults',
                default={'allocation_ratio': 1.0,
                         'min_unit': 1,
                         'step_size': 1,
                         'reserved': 0},
                help=_("Key:value pairs to specify defaults used "
                       "while reporting resource provider inventories. "
                       "Possible keys with their types: "
                       "allocation_ratio:float, "
                       "max_unit:int, min_unit:int, "
                       "reserved:int, step_size:int, "
                       "See also: "
                       "https://docs.openstack.org/api-ref/placement/"
                       "#update-resource-provider-inventories")),
]


def register_agent_sriov_nic_opts(cfg=cfg.CONF):
    cfg.register_opts(sriov_nic_opts, 'SRIOV_NIC')



# Copyright 2016 Red Hat, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
import testtools

from neutron.plugins.ml2.drivers.openvswitch.agent import vlanmanager
from neutron.tests import base


class LocalVlanManagerFixture(fixtures.Fixture):
    def _setUp(self):
        super(LocalVlanManagerFixture, self)._setUp()
        self.vlan_manager = vlanmanager.LocalVlanManager()
        self.addCleanup(self.restore_manager)
        # Remove _instance attribute from VlanManager in order to not obtain a
        # singleton
        del vlanmanager.LocalVlanManager._instance
        self.manager = vlanmanager.LocalVlanManager()

    def restore_manager(self):
        vlanmanager.LocalVlanManager._instance = self.vlan_manager


class TestLocalVLANMapping(base.BaseTestCase):
    def test___eq___equal(self):
        mapping1 = vlanmanager.LocalVLANMapping(1, 2, 3, 4, 5)
        mapping2 = vlanmanager.LocalVLANMapping(1, 2, 3, 4, 5)
        self.assertEqual(mapping1, mapping2)

    def test___eq___different(self):
        mapping1 = vlanmanager.LocalVLANMapping(1, 2, 3, 4, 5)
        mapping2 = vlanmanager.LocalVLANMapping(1, 2, 4, 4, 5)
        self.assertNotEqual(mapping1, mapping2)

    def test___eq___different_type(self):
        mapping = vlanmanager.LocalVLANMapping(1, 2, 3, 4, 5)
        self.assertNotEqual(mapping, "foo")


class TestLocalVlanManager(base.BaseTestCase):

    def setUp(self):
        super(TestLocalVlanManager, self).setUp()
        self.vlan_manager = self.useFixture(LocalVlanManagerFixture()).manager

    def test_is_singleton(self):
        self.vlan_manager.add(1, None, None, None, None)
        new_vlan_manager = vlanmanager.LocalVlanManager()
        self.assertIs(new_vlan_manager, self.vlan_manager)
        self.assertCountEqual(new_vlan_manager.mapping,
                              self.vlan_manager.mapping)

    def test_in_operator_on_key(self):
        self.vlan_manager.add(1, None, None, None, None)
        self.assertIn(1, self.vlan_manager)
        self.assertNotIn(2, self.vlan_manager)

    def test_iterator_returns_vlan_mappings(self):
        created_vlans = []
        for val in range(3):
            self.vlan_manager.add(val, val, val, val, val)
            created_vlans.append({val: self.vlan_manager.get(val, val)})

        self.assertCountEqual(created_vlans, list(self.vlan_manager))

    def test_get_net_and_segmentation_id_existing(self):
        port_id = 'port-id'
        vlan_data = (2, 3, 4, 5, {port_id: 'port'})
        net_id = 1
        self.vlan_manager.add(net_id, *vlan_data)
        obtained_net_id = (
            self.vlan_manager.get_net_and_segmentation_id(port_id))
        self.assertEqual((net_id, 5), obtained_net_id)

    def test_get_net_and_segmentation_id_non_existing_raises_exception(self):
        vlan_data = (1, 2, 3, 4, 5, {'port_id': 'port'})
        self.vlan_manager.add(*vlan_data)
        with testtools.ExpectedException(vlanmanager.VifIdNotFound):
            self.vlan_manager.get_net_and_segmentation_id('non-existing-port')

    def test_add_and_get(self):
        vlan_data = (2, 3, 4, 5, 6)
        expected_vlan_mapping = vlanmanager.LocalVLANMapping(*vlan_data)
        self.vlan_manager.add(1, *vlan_data)
        vlan_mapping = self.vlan_manager.get(1, 5)
        self.assertEqual(expected_vlan_mapping, vlan_mapping)

    def test_add_existing_raises_exception(self):
        vlan_data = (2, 3, 4, 5, 6)
        self.vlan_manager.add(1, *vlan_data)
        with testtools.ExpectedException(vlanmanager.MappingAlreadyExists):
            self.vlan_manager.add(1, *vlan_data)

    def test_get_non_existing_raises_keyerror(self):
        with testtools.ExpectedException(vlanmanager.MappingNotFound):
            self.vlan_manager.get(1, 5)

    def test_pop(self):
        vlan_data = (2, 3, 4, 5, 6)
        expected_vlan_mapping = vlanmanager.LocalVLANMapping(*vlan_data)
        self.vlan_manager.add(1, *vlan_data)
        vlan_mapping = self.vlan_manager.pop(1, 5)
        self.assertEqual(expected_vlan_mapping, vlan_mapping)
        self.assertFalse(self.vlan_manager.mapping)

    def test_pop_non_existing_raises_exception(self):
        with testtools.ExpectedException(vlanmanager.MappingNotFound):
            self.vlan_manager.pop(1, 5)

    def test_update_segmentation_id(self):
        self.vlan_manager.add('net_id', 'vlan_id', 'vlan', 'phys_net',
                              1001, None)
        self.assertEqual(1001, self.vlan_manager.get(
            'net_id', 1001).segmentation_id)
        self.vlan_manager.update_segmentation_id('net_id', 1002)
        self.assertEqual(1002, self.vlan_manager.get(
            'net_id', 1002).segmentation_id)

    def test_update_segmentation_id_not_found(self):
        with testtools.ExpectedException(vlanmanager.MappingNotFound):
            self.vlan_manager.update_segmentation_id(
                'net_id-notfound', 1002)

    def test_update_segmentation_id_not_uniq(self):
        self.vlan_manager.add('net_id-not-uniq', 'vlan_id', 'vlan', 'phys_net',
                              1001, None)
        self.vlan_manager.add('net_id-not-uniq', 'vlan_id', 'vlan', 'phys_net',
                              1002, None)
        with testtools.ExpectedException(vlanmanager.NotUniqMapping):
            self.vlan_manager.update_segmentation_id(
                'net_id-not-uniq', 1003)
